# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""Tests for editor decorations."""

# Third party imports
import os.path as osp
import random
from unittest.mock import patch

from flaky import flaky
import pytest
from qtpy.QtCore import Qt
from qtpy.QtGui import QFont, QTextCursor

# Local imports
from spyder.plugins.editor.widgets.codeeditor import CodeEditor


HERE = osp.dirname(osp.realpath(__file__))
PARENT = osp.dirname(HERE)


def test_decorations(codeeditor, qtbot):
    """Test decorations."""
    editor = codeeditor

    # Set random size
    editor.resize(640, random.randint(200, 500))

    # Set cell of different length.
    base_function = (
        "def some_function():\n"
        "    some_variable = 1\n"
        "    some_variable += 2\n"
        "    return some_variable\n\n"
    )

    text = ''
    for __ in range(100):
        base_text = base_function * random.randint(2, 8) + "# %%\n"
        text = text + base_text

    editor.set_text(text)

    # Move cursor over 'some_variable'
    editor.go_to_line(2)
    cursor = editor.textCursor()
    cursor.movePosition(QTextCursor.Right, n=5)
    editor.setTextCursor(cursor)

    # Assert number of decorations is the one we expect.
    qtbot.wait(3000)
    decorations = editor.decorations._sorted_decorations()

    assert len(decorations) == 2 + text.count('some_variable')

    # Assert that selection 0 is current cell
    assert decorations[0].kind == 'current_cell'

    # Assert that selection 1 is current_line
    assert decorations[1].kind == 'current_line'

    # Assert the other decorations are occurrences
    assert all([d.kind == 'occurrences' for d in decorations[2:5]])

    # Assert all other decorations are some_variable
    selected_texts = [d.cursor.selectedText() for d in decorations]
    assert set(selected_texts[2:]) == set(['some_variable'])

    # Assert painted extra selections are much smaller.
    first, last = editor.get_buffer_block_numbers()
    max_decorations = last - first
    assert len(editor.extraSelections()) < max_decorations

    # Clear decorations to be sure they are painted again below.
    editor.decorations.clear()
    editor.decorations._update()
    assert editor.decorations._sorted_decorations() == []

    # Move to a random place in the file and wait until decorations are
    # updated.
    line_number = random.randint(100, editor.blockCount())
    editor.go_to_line(line_number)
    qtbot.wait(editor.UPDATE_DECORATIONS_TIMEOUT + 100)

    # Assert a new cell is painted
    decorations = editor.decorations._sorted_decorations()
    assert decorations[0].kind == 'current_cell'


@flaky(max_runs=10)
def test_update_decorations_when_scrolling(qtbot):
    """
    Test how many calls we're doing to update decorations when
    scrolling.
    """
    # NOTE: Here we need to use `patch` from unittest.mock, instead of the
    # mocker fixture, to have the same results when running the test
    # alone and with the other tests in this file.

    patched_object = ('spyder.plugins.editor.utils.decoration.'
                      'TextDecorationsManager._update')

    with patch(patched_object) as _update:
        # NOTE: We can't use a fixture to build a CodeEditor instance here
        # because the testing results are not consistent.
        editor = CodeEditor(parent=None)
        editor.setup_editor(
            language='Python',
            color_scheme='spyder/dark',
            font=QFont("Monospace", 10),
        )
        editor.resize(640, 480)
        editor.show()
        qtbot.addWidget(editor)

        # If there's no waiting after CodeEditor is created, there shouldn't
        # be a call to _update.
        assert _update.call_count == 0

        with open(osp.join(PARENT, 'codeeditor.py'), 'r', encoding='utf-8') as f:
            text = f.read()
        editor.set_text(text)

        # If there's no waiting after setting text, there shouldn't be a
        # call to _update either.
        assert _update.call_count == 0

        # Simulate scrolling
        scrollbar = editor.verticalScrollBar()
        for i in range(6):
            scrollbar.setValue(i * 70)
            qtbot.wait(100)

        # A new call is done here due to __cursor_position_changed being
        # called, which in turn calls highlight_current_cell and
        # highlight_current_line
        assert _update.call_count == 1

        # Wait for decorations to update
        qtbot.wait(editor.UPDATE_DECORATIONS_TIMEOUT + 100)

        # Assert a new call to _update was done
        assert _update.call_count == 2

        # Simulate grabbing and moving the scrollbar with the mouse
        scrollbar = editor.verticalScrollBar()
        value = scrollbar.value()
        for __ in range(400):
            scrollbar.setValue(value + 1)
            value = scrollbar.value()

        # No calls should be done after this.
        assert _update.call_count == 2

        # Wait for decorations to update
        qtbot.wait(editor.UPDATE_DECORATIONS_TIMEOUT + 100)

        # Assert a new call to _update was done
        assert _update.call_count == 3

        # Move to the last visible line
        _, last = editor.get_visible_block_numbers()
        editor.go_to_line(last)

        # Simulate continuously pressing the down arrow key.
        for __ in range(200):
            qtbot.keyPress(editor, Qt.Key_Down)

        # Only one call to _update should be done, after releasing the key.
        qtbot.wait(editor.UPDATE_DECORATIONS_TIMEOUT + 100)
        assert _update.call_count == 5

        # Simulate continuously pressing the up arrow key.
        for __ in range(200):
            qtbot.keyPress(editor, Qt.Key_Up)

        # Only one call to _update should be done, after releasing the key.
        qtbot.wait(editor.UPDATE_DECORATIONS_TIMEOUT + 100)
        assert _update.call_count == 6


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Status bar widgets."""

# Third party imports
import qstylizer.parser
from qtpy import PYQT5, PYQT6
from qtpy.QtCore import Qt, QSize, QTimer, Signal
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QHBoxLayout, QLabel, QWidget

# Local imports
from spyder.api.exceptions import SpyderAPIError
from spyder.api.widgets.mixins import SpyderWidgetMixin
from spyder.utils.palette import SpyderPalette
from spyder.utils.qthelpers import create_waitspinner
from spyder.utils.stylesheet import MAC


class StatusBarWidget(QWidget, SpyderWidgetMixin):
    """
    Base class for status bar widgets.

    These widgets consist by default of an icon, a label and a spinner,
    which are organized from left to right on that order.

    You can also add any other QWidget to this layout by setting the
    CUSTOM_WIDGET_CLASS class attribute. It'll be put between the label
    and the spinner.
    """

    ID = None
    """
    Unique string widget identifier.
    """

    CUSTOM_WIDGET_CLASS = None
    """
    Custom widget class to add to the default layout.
    """

    INTERACT_ON_CLICK = False
    """
    Whether the user can interact with widget when clicked (e.g. to show a
    menu)
    """

    sig_clicked = Signal()
    """
    This signal is emmitted when the widget is clicked.
    """

    def __init__(self, parent=None, show_icon=True, show_label=True,
                 show_spinner=False):
        """
        Base class for status bar widgets.

        These are composed of the following widgets, which are arranged
        in a QHBoxLayout from left to right:

        * Icon
        * Label
        * Custom QWidget
        * Spinner

        Parameters
        ----------
        show_icon: bool
            Show an icon in the widget.
        show_label: bool
            Show a label in the widget.
        show_spinner: bool
            Show a spinner.

        Notes
        -----
        1. To use an icon, you need to redefine the ``get_icon`` method.
        2. To use a label, you need to call ``set_value``.
        """
        if PYQT5 or PYQT6:
            super().__init__(parent, class_parent=parent)
        else:
            QWidget.__init__(self, parent)
            SpyderWidgetMixin.__init__(self, class_parent=parent)

        self._parent = parent

        self.show_icon = show_icon
        self.show_label = show_label
        self.show_spinner = show_spinner

        self.value = None
        self.label_icon = None
        self.label_value = None
        self.spinner = None
        self.custom_widget = None

        # In case the widget has an associated menu
        self.menu = None

        self._set_layout()
        self._css = self._generate_stylesheet()
        self.setStyleSheet(self._css.toString())

    # ---- Private API
    # -------------------------------------------------------------------------
    def _set_layout(self):
        """Set layout for default widgets."""
        # Icon
        if self.show_icon:
            self._icon = self.get_icon()
            self._pixmap = None
            self._icon_size = QSize(16, 16)  # Should this be adjustable?
            self.label_icon = QLabel(self)
            self.set_icon()

        # Label
        if self.show_label:
            self.label_value = QLabel(self)
            self.set_value('')
            self.label_value.setAlignment(Qt.AlignRight | Qt.AlignVCenter)

        # Custom widget
        if self.CUSTOM_WIDGET_CLASS:
            if not issubclass(self.CUSTOM_WIDGET_CLASS, QWidget):
                raise SpyderAPIError(
                    'Any custom status widget must subclass QWidget!'
                )
            self.custom_widget = self.CUSTOM_WIDGET_CLASS(self._parent)

        # Spinner
        if self.show_spinner:
            self.spinner = create_waitspinner(size=14, parent=self)
            self.spinner.hide()

        # Layout setup
        layout = QHBoxLayout(self)
        layout.setSpacing(0)  # Reduce space between icon and label
        if self.show_icon:
            layout.addWidget(self.label_icon)
        if self.show_label:
            layout.addWidget(self.label_value)
        if self.custom_widget:
            layout.addWidget(self.custom_widget)
        if self.show_spinner:
            layout.addWidget(self.spinner)

        layout.addSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setAlignment(Qt.AlignVCenter)

        # Setup
        self.update_tooltip()

    def _generate_stylesheet(self):
        """Generate the widget's stylesheet."""
        # Remove opacity that comes from QDarkstyle.
        # This work around is necessary because qstylizer doesn't have support
        # for the opacity property.
        initial_css = "QToolTip {opacity: 255;}"
        css = qstylizer.parser.parse(initial_css)

        # Make style match the one set for other tooltips in the app
        css.QToolTip.setValues(
            color=SpyderPalette.COLOR_TEXT_1,
            backgroundColor=SpyderPalette.COLOR_ACCENT_2,
            border="none",
            padding="1px 2px",
        )

        return css

    # ---- Public API
    # -------------------------------------------------------------------------
    def get_icon(self):
        """Get the widget's icon."""
        return None

    def set_icon(self):
        """Set the icon for the status bar widget."""
        if self.label_icon:
            icon = self._icon
            self.label_icon.setVisible(icon is not None)
            if icon is not None and isinstance(icon, QIcon):
                self._pixmap = icon.pixmap(self._icon_size)
                self.label_icon.setPixmap(self._pixmap)

    def set_value(self, value):
        """Set formatted text value."""
        if self.label_value:
            self.value = value
            self.label_value.setText(value)

    def get_tooltip(self):
        """Get the widget's tooltip text."""
        return ''

    def update_tooltip(self):
        """Update tooltip for widget."""
        tooltip = self.get_tooltip()
        if tooltip:
            if self.label_value:
                self.label_value.setToolTip(tooltip)
            if self.label_icon:
                self.label_icon.setToolTip(tooltip)
            self.setToolTip(tooltip)

    # ---- Qt methods
    # -------------------------------------------------------------------------
    def mousePressEvent(self, event):
        """Change background color when the widget is clicked."""
        if self.INTERACT_ON_CLICK:
            self._css.QWidget.setValues(
                backgroundColor=SpyderPalette.COLOR_BACKGROUND_6
            )
            self.setStyleSheet(self._css.toString())

        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        """
        Change background color and emit signal to inform the widget was
        clicked.
        """
        super().mouseReleaseEvent(event)

        if self.INTERACT_ON_CLICK:
            self._css.QWidget.setValues(
                # Mac doesn't correctly restore the background color after
                # clicking on a widget that shows a menu
                backgroundColor=SpyderPalette.COLOR_BACKGROUND_4
                if MAC and self.menu
                else SpyderPalette.COLOR_BACKGROUND_5
            )
            self.setStyleSheet(self._css.toString())

        self.sig_clicked.emit()

    def enterEvent(self, event):
        """Change background color and cursor shape on hover."""
        if self.INTERACT_ON_CLICK:
            self._css.QWidget.setValues(
                backgroundColor=SpyderPalette.COLOR_BACKGROUND_5
            )
            self.setStyleSheet(self._css.toString())

            self.setCursor(Qt.PointingHandCursor)

        self.update_tooltip()
        super().enterEvent(event)

    def leaveEvent(self, event):
        """Restore background color when not hovering."""
        if self.INTERACT_ON_CLICK:
            self._css.QWidget.setValues(
                backgroundColor=SpyderPalette.COLOR_BACKGROUND_4
            )
            self.setStyleSheet(self._css.toString())

        super().leaveEvent(event)


class BaseTimerStatus(StatusBarWidget):
    """
    Base class for status bar widgets that update based on timers.
    """

    def __init__(self, parent=None):
        """Base class for status bar widgets that update based on timers."""
        self.timer = None  # Needs to come before parent call
        super().__init__(parent)
        self._interval = 2000

        # Widget setup
        fm = self.label_value.fontMetrics()
        self.label_value.setMinimumWidth(fm.width('000%'))

        # Setup
        self.timer = QTimer(self)
        self.timer.timeout.connect(self.update_status)
        self.timer.start(self._interval)

    # ---- Qt methods
    # -------------------------------------------------------------------------
    def closeEvent(self, event):
        self.timer.stop()
        super().closeEvent(event)

    def setVisible(self, value):
        """Stop timer if widget is not visible."""
        if self.timer is not None:
            if value:
                self.timer.start(self._interval)
            else:
                self.timer.stop()
        super().setVisible(value)

    # ---- Public API
    # -------------------------------------------------------------------------
    def update_status(self):
        """Update status label widget, if widget is visible."""
        if self.isVisible():
            self.label_value.setText(self.get_value())

    def set_interval(self, interval):
        """Set timer interval (ms)."""
        self._interval = interval
        if self.timer is not None:
            self.timer.setInterval(interval)

    def get_value(self):
        """Return formatted text value."""
        raise NotImplementedError



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# ----------------------------------------------------------------------------
"""Tests for plugin config dialog."""

# Standard library imports
import pytest

# Local imports
from spyder.plugins.profiler.plugin import Profiler
from spyder.plugins.preferences.tests.conftest import config_dialog


@pytest.mark.parametrize(
    'config_dialog',
    [[None, [], [Profiler]]],
    indirect=True)
def test_config_dialog(config_dialog):
    configpage = config_dialog.get_page()
    configpage.save_to_conf()
    assert configpage



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Language Server Protocol configuration tabs.
"""

# Standard library imports
import re

# Third party imports
from qtpy.QtCore import Qt, Slot
from qtpy.QtWidgets import (QGridLayout, QLabel, QMessageBox, QVBoxLayout,
                            QWidget)

# Local imports
from spyder.api.preferences import SpyderPreferencesTab
from spyder.config.base import _


class DocstringConfigTab(SpyderPreferencesTab):
    """Docstring style configuration tab."""

    TITLE = _('Docstring style')

    def __init__(self, parent):
        super().__init__(parent)

        numpy_url = (
            "<a href='https://numpydoc.readthedocs.io/en/"
            "latest/format.html'>Numpy</a>")
        pep257_url = (
            "<a href='https://www.python.org/dev/peps/pep-0257/'>PEP 257</a>")
        docstring_style_codes = _(
            "<a href='http://www.pydocstyle.org/en/stable"
            "/error_codes.html'>page</a>")
        docstring_style_label = QLabel(
            _("Here you can decide if you want to perform style analysis on "
              "your docstrings according to the {} or {} conventions. You can "
              "also decide if you want to show or ignore specific errors, "
              "according to the codes found on this {}.").format(
                  numpy_url, pep257_url, docstring_style_codes))
        docstring_style_label.setOpenExternalLinks(True)
        docstring_style_label.setWordWrap(True)

        # Docstring style checkbox
        self.docstring_style_check = self.create_checkbox(
            _("Enable docstring style linting"),
            'pydocstyle')

        # Docstring style options
        docstring_style_convention = self.create_combobox(
            _("Choose the convention used to lint docstrings: "),
            (("Numpy", 'numpy'),
             ("PEP 257", 'pep257'),
             ("Custom", 'custom')),
            'pydocstyle/convention')
        self.docstring_style_select = self.create_lineedit(
            _("Show the following errors:"),
            'pydocstyle/select', alignment=Qt.Horizontal, word_wrap=False,
            placeholder=_("Example codes: D413, D414"))
        self.docstring_style_ignore = self.create_lineedit(
            _("Ignore the following errors:"),
            'pydocstyle/ignore', alignment=Qt.Horizontal, word_wrap=False,
            placeholder=_("Example codes: D107, D402"))
        self.docstring_style_match = self.create_lineedit(
            _("Only check filenames matching these patterns:"),
            'pydocstyle/match', alignment=Qt.Horizontal, word_wrap=False,
            placeholder=_("Skip test files: (?!test_).*\\.py"))
        self.docstring_style_match_dir = self.create_lineedit(
            _("Only check in directories matching these patterns:"),
            'pydocstyle/match_dir', alignment=Qt.Horizontal, word_wrap=False,
            placeholder=_("Skip dot directories: [^\\.].*"))

        # Custom option handling
        docstring_style_convention.combobox.currentTextChanged.connect(
                self.setup_docstring_style_convention)
        current_convention = docstring_style_convention.combobox.currentText()
        self.setup_docstring_style_convention(current_convention)

        # Docstring style layout
        docstring_style_g_layout = QGridLayout()
        docstring_style_g_layout.addWidget(
            docstring_style_convention.label, 1, 0)
        docstring_style_g_layout.addWidget(
            docstring_style_convention.combobox, 1, 1)
        docstring_style_g_layout.addWidget(
            self.docstring_style_select.label, 2, 0)
        docstring_style_g_layout.addWidget(
            self.docstring_style_select.textbox, 2, 1)
        docstring_style_g_layout.addWidget(
            self.docstring_style_ignore.label, 3, 0)
        docstring_style_g_layout.addWidget(
            self.docstring_style_ignore.textbox, 3, 1)
        docstring_style_g_layout.addWidget(
            self.docstring_style_match.label, 4, 0)
        docstring_style_g_layout.addWidget(
            self.docstring_style_match.textbox, 4, 1)
        docstring_style_g_layout.addWidget(
            self.docstring_style_match_dir.label, 5, 0)
        docstring_style_g_layout.addWidget(
            self.docstring_style_match_dir.textbox, 5, 1)

        # Set Docstring style options enabled/disabled
        docstring_style_g_widget = QWidget()
        docstring_style_g_widget.setLayout(docstring_style_g_layout)
        docstring_style_g_widget.setEnabled(self.get_option('pydocstyle'))
        self.docstring_style_check.checkbox.toggled.connect(
            docstring_style_g_widget.setEnabled)

        # Docstring style layout
        docstring_style_layout = QVBoxLayout()
        docstring_style_layout.addWidget(docstring_style_label)
        docstring_style_layout.addWidget(self.docstring_style_check)
        docstring_style_layout.addWidget(docstring_style_g_widget)

        self.setLayout(docstring_style_layout)

    @Slot(str)
    def setup_docstring_style_convention(self, text):
        """Handle convention changes."""
        if text == 'Custom':
            self.docstring_style_select.label.setText(
                _("Show the following errors:"))
            self.docstring_style_ignore.label.setText(
                _("Ignore the following errors:"))
        else:
            self.docstring_style_select.label.setText(
                _("Show the following errors in addition "
                  "to the specified convention:"))
            self.docstring_style_ignore.label.setText(
                _("Ignore the following errors in addition "
                  "to the specified convention:"))

    def report_invalid_regex(self, files=True):
        """
        Report that matching files/directories should be valid regular
        expressions.
        """
        msg = _('Directory patterns listed for matching should be valid '
                'regular expressions')
        if files:
            msg = _('File patterns listed for matching should be valid '
                    'regular expressions')

        QMessageBox.critical(self, _("Error"), msg)

    def is_valid(self):
        # Check regex of docstring style options
        try:
            docstring_style_match = (
                self.docstring_style_match.textbox.text())
            re.compile(docstring_style_match)
        except re.error:
            self.report_invalid_regex()
            return False

        try:
            docstring_style_match_dir = (
                self.docstring_style_match_dir.textbox.text())
            re.compile(docstring_style_match_dir)
        except re.error:
            self.report_invalid_regex(files=False)
            return False
        return True



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Explorer widget utilities."""

# Standard library imports
import os
import os.path as osp
import re
import subprocess
import sys

# Third-party imports
from qtpy.QtCore import QFileInfo, Slot
from qtpy.QtWidgets import QFileIconProvider, QMessageBox

# Local imports
from spyder.api.translations import _
from spyder.utils import encoding
from spyder.utils.icon_manager import ima


def open_file_in_external_explorer(filename):
    if sys.platform == "darwin":
        subprocess.call(["open", "-R", filename])
    elif os.name == 'nt':
        if os.path.exists(filename):
            subprocess.call(["explorer", "/select,", filename])
        else:
            subprocess.call(["explorer", os.path.dirname(filename)])
    else:
        filename = os.path.dirname(filename)
        subprocess.call(["xdg-open", filename])


def show_in_external_file_explorer(fnames=None):
    """Show files in external file explorer

    Args:
        fnames (list): Names of files to show.
    """
    if not isinstance(fnames, (tuple, list)):
        fnames = [fnames]
    for fname in fnames:
        open_file_in_external_explorer(fname)


def fixpath(path):
    """Normalize path fixing case, making absolute and removing symlinks"""
    norm = osp.normcase if os.name == 'nt' else osp.normpath
    return norm(osp.abspath(osp.realpath(path)))


def create_script(fname):
    """Create a new Python script"""
    text = os.linesep.join(["# -*- coding: utf-8 -*-", "", ""])
    try:
        encoding.write(str(text), fname, 'utf-8')
    except EnvironmentError as error:
        QMessageBox.critical(_("Save Error"),
                             _("<b>Unable to save file '%s'</b>"
                               "<br><br>Error message:<br>%s"
                               ) % (osp.basename(fname), str(error)))


def listdir(path, include=r'.', exclude=r'\.pyc$|^\.', folders_only=False):
    """List files and directories"""
    namelist = []
    dirlist = [str(osp.pardir)]
    for item in os.listdir(str(path)):
        if re.search(exclude, item):
            continue
        if osp.isdir(osp.join(path, item)):
            dirlist.append(item)
        elif folders_only:
            continue
        elif re.search(include, item):
            namelist.append(item)
    return sorted(dirlist, key=str.lower) + sorted(namelist, key=str.lower)


def has_subdirectories(path, include, exclude):
    """Return True if path has subdirectories"""
    try:
        # > 1 because of '..'
        return len(listdir(path, include, exclude, folders_only=True)) > 1
    except (IOError, OSError):
        return False


class IconProvider(QFileIconProvider):
    """Project tree widget icon provider"""

    @Slot(int)
    @Slot(QFileInfo)
    def icon(self, icontype_or_qfileinfo):
        """Reimplement Qt method"""
        if isinstance(icontype_or_qfileinfo, QFileIconProvider.IconType):
            return super().icon(icontype_or_qfileinfo)
        else:
            qfileinfo = icontype_or_qfileinfo
            fname = osp.normpath(str(qfileinfo.absoluteFilePath()))

            if osp.isfile(fname) or osp.isdir(fname):
                icon = ima.get_icon_by_extension_or_type(
                    fname, scale_factor=1.0
                )
            else:
                icon = ima.icon('binary')

            return icon



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Tests for config/user.py."""

# Standard library imports
import configparser as cp
import os

# Third party imports
import pytest

# Local imports
from spyder.config.main import CONF_VERSION, DEFAULTS
from spyder.config.user import NoDefault, UserConfig
from spyder.utils.fixtures import tmpconfig


# --- Default config tests
# ----------------------------------------------------------------------------
@pytest.mark.parametrize(
    "test_input, expected",
    [
        (('sec', 'opt', 'val'), '[sec][opt] = val\n'),
        (('sec', 'opt', 50), '[sec][opt] = 50\n'),
        (('sec', 'opt', [50]), '[sec][opt] = [50]\n'),
        (('sec', 'opt', (50, 2)), '[sec][opt] = (50, 2)\n'),
        (('sec', 'opt', {50}), '[sec][opt] = {50}\n'),
        (('sec', 'opt', {'k': 50}), "[sec][opt] = {'k': 50}\n"),
        (('sec', 'opt', False), '[sec][opt] = False\n'),
        (('sec', 'opt', True), '[sec][opt] = True\n'),
        (('sec space', 'opt', True), '[sec space][opt] = True\n'),
        (('sec space', 'opt space', True), '[sec space][opt space] = True\n'),
    ]
)
def test_default_config_set(defaultconfig, capsys, test_input, expected):
    section, option, value = test_input
    defaultconfig._set(section, option, value, verbose=True)
    assert defaultconfig.sections() == [section]

    captured = capsys.readouterr()
    assert captured.out == expected


def test_default_config_save_write(defaultconfig):
    defaultconfig._save()


def test_default_config_set_defaults(defaultconfig):
    defaults = [('main2', {'opt': 1})]
    defaultconfig.set_defaults(defaults)
    assert 'main2' in defaultconfig.sections()


# --- User config tests
# ============================================================================
# --- Helpers and checkers
@pytest.mark.parametrize('value,expected', [
    ('3.2.1', '3.2'),
    ('3.2.0', '3.2'),
])
def test_userconfig_get_minor_version(value, expected):
    result = UserConfig._get_minor_version(value)
    assert result == expected


@pytest.mark.parametrize('value,expected', [
    ('3.2.1', '3'),
    ('0.2.0', '0'),
])
def test_userconfig_get_major_version(value, expected):
    result = UserConfig._get_major_version(value)
    assert result == expected


@pytest.mark.parametrize('test_version', [
    ('abc'),
    ('x.x.x'),
    ('1.0'),
    ('-1.0'),
    ('-1.0.0'),
    (''),
])
def test_userconfig_check_version(tmpdir, test_version, userconfig):
    name = 'spyder-test'
    path = str(tmpdir)
    with pytest.raises(ValueError):
        UserConfig(name=name, path=path, defaults=DEFAULTS,
                   load=False, version=test_version, raw_mode=True)

    # Check that setting a wrong version also runs the version checks
    with pytest.raises(ValueError):
        userconfig.set_version(test_version)


def test_userconfig_check_defaults(tmpdir, capsys):
    name = 'foobar'
    path = str(tmpdir)
    conf = UserConfig(name=name, path=path, defaults={},
                      load=False, version='1.0.0', backup=False,
                      raw_mode=True)

    conf._check_defaults({})
    conf._check_defaults({'option2': 'value2'})
    conf._check_defaults([])
    conf._check_defaults([('sec', {'opt1': 'val1'})])

    with pytest.raises(AssertionError):
        conf._check_defaults([(123, {'opt1': 'val1'})])

    with pytest.raises(AssertionError):
        conf._check_defaults([('sec', {123: 'val1'})])

    with pytest.raises(ValueError):
        conf._check_defaults({1, 2, 3})

    with pytest.raises(ValueError):
        conf._check_defaults('asd')


def test_userconfig_check_section_option(userconfig):
    section = userconfig._check_section_option(None, 'version')
    assert section == userconfig.DEFAULT_SECTION_NAME

    section = userconfig._check_section_option(None, 'opt')
    assert section == userconfig.DEFAULT_SECTION_NAME

    section = userconfig._check_section_option('sec', 'opt')
    assert section == 'sec'

    # Check with invalid section
    with pytest.raises(RuntimeError):
        section = userconfig._check_section_option(123, 'opt')

    # Check with invalid option
    with pytest.raises(RuntimeError):
        section = userconfig._check_section_option(None, 123)


def test_userconfig_load_from_ini(tmpdir, capsys):
    # Test error on loading
    name = 'foobar'
    path = str(tmpdir)
    fpath = os.path.join(path, '{}.ini'.format(name))
    with open(fpath, 'w') as fh:
        fh.write('[sec\n')

    conf = UserConfig(name=name, path=path, defaults=DEFAULTS,
                      load=True, version=CONF_VERSION, backup=False,
                      raw_mode=True)
    assert conf.get_config_fpath() == fpath

    captured = capsys.readouterr()
    assert 'Warning' in captured.out


# --- Public API
def test_userconfig_get_version(userconfig, tmpconfig):
    assert tmpconfig.get_version() == CONF_VERSION
    assert userconfig.get_version() == '1.0.0'
    userconfig.remove_option('main', 'version')
    assert userconfig.get_version() == '0.0.0'


def test_userconfig_set_version(userconfig):
    version = '1000.1000.1000'
    userconfig.set_version(version)
    assert userconfig.get_version() == version


def test_userconfig_reset_to_defaults(tmpdir):
    name = 'foobar'
    path = str(tmpdir)
    defaults = [('main', {'opt': False}), ('test', {'opt': False})]
    conf = UserConfig(name=name, path=path, defaults=defaults,
                      load=False, version='1.0.0', backup=False,
                      raw_mode=True)
    # Skip section, should go to default
    assert conf.defaults == defaults
    conf.set(None, 'opt', True)
    assert conf.get(None, 'opt') is True
    conf.reset_to_defaults()
    assert conf.get(None, 'opt') is False

    # Provide section, should go to sectio
    assert conf.defaults == defaults
    conf.set('test', 'opt', True)
    assert conf.get('test', 'opt') is True
    conf.reset_to_defaults()
    assert conf.get('test', 'opt') is False


def test_userconfig_set_as_defaults(tmpdir):
    name = 'foobar'
    path = str(tmpdir)
    conf = UserConfig(name=name, path=path, defaults={},
                      load=False, version='1.0.0', backup=False,
                      raw_mode=True)

    # TODO: Is this expected? this seems inconsistent
    assert conf.defaults == [('main', {})]
    conf.set_as_defaults()
    assert conf.defaults == []


def test_userconfig_get_default(userconfig, tmpconfig):
    # Not existing and no defaults
    value = userconfig.get_default('other_section', 'other_option')
    assert value == NoDefault

    # Existing and no defaults
    value = userconfig.get_default('section', 'option')
    assert value == NoDefault

    # Existing and defaults
    value = tmpconfig.get_default('main', 'window/is_maximized')
    assert value is True


class TestUserConfigGet:

    @pytest.mark.parametrize(
        'defaults,value',
        [
            # Valid values
            ([('test', {'opt': 'value'})], 'value'),
            ([('test', {'opt': u'"éàÇÃãéèï"'})], u'"éàÇÃãéèï"'),
            ([('test', {'opt': 'éàÇÃãéèï'})], u'éàÇÃãéèï'),
            ([('test', {'opt': True})], True),
            ([('test', {'opt': UserConfig})], repr(UserConfig)),
            ([('test', {'opt': 123})], 123),
            ([('test', {'opt': 123.123})], 123.123),
            ([('test', {'opt': [1]})], [1]),
            ([('test', {'opt': {'key': 'val'}})], {'key': 'val'}),
        ]
    )
    def test_userconfig_get(self, defaults, value, tmpdir):
        name = 'foobar'
        path = str(tmpdir)
        conf = UserConfig(name=name, path=path, defaults=defaults,
                          load=False, version='1.0.0', backup=False,
                          raw_mode=True)

        assert conf.get('test', 'opt') == value

    @pytest.mark.parametrize(
        'defaults,default,raises',
        [
            # Valid values
            ([('test2', {'opt': 'value'})], 'val', True),
            ([('test2', {'opt': 'value'})], 'val', False),
            ([('test', {'opt': 'value'})], 'val', False),
        ]
    )
    def test_userconfig_get2(self, defaults, default, raises, tmpdir):
        name = 'foobar'
        path = str(tmpdir)
        conf = UserConfig(name=name, path=path, defaults=defaults,
                          load=False, version='1.0.0', backup=False,
                          raw_mode=True)

        if raises:
            with pytest.raises(cp.NoSectionError):
                conf.get('test', 'opt')
        else:
            conf.get('test', 'opt', default)

    def test_userconfig_get_string_from_inifile(self, userconfig):
        assert userconfig.get('section', 'option') == 'value'

    def test_userconfig_get_does_not_eval_functions(self, userconfig):
        # Regression test for spyder-ide/spyder#3354.
        userconfig.set('section', 'option', 'print("foo")')
        assert userconfig.get('section', 'option') == 'print("foo")'


def test_userconfig_set_default(userconfig):
    value = userconfig.get_default('section', 'option')
    assert value == NoDefault

    # TODO: Is this expected behavior? should. If no defaults are provided
    # then they cannot be set individually until set_as_defaults() is ran?
    default_value = 'foobar'
    value = userconfig.set_default('section', 'option', default_value)
    value = userconfig.get_default('section', 'option')
    assert value == NoDefault

    userconfig.set_as_defaults()
    value = userconfig.get_default('section', 'option')
    expected = 'value'
    assert value == expected
    value = userconfig.set_default('section', 'option', default_value)
    value = userconfig.get_default('section', 'option')
    assert value == default_value


class TestUserConfigSet:

    @pytest.mark.parametrize(
        'defaults,value',
        [
            # Valid values
            ([('test', {'opt': 'value'})], 'other'),
            ([('test', {'opt': 'éàÇÃãéèï'})], u'ãéèï'),
            ([('test', {'opt': True})], False),
            ([('test', {'opt': UserConfig})], dict),
            ([('test', {'opt': 123})], 345),
            ([('test', {'opt': 123.123})], 345.345),
            ([('test', {'opt': [1]})], [1]),
            ([('test', {'opt': {'key': 'val'}})], {'key2': 'val2'}),
            # Value not in defaults
            ([('test', {'opt1': 'value'})], 'other'),
        ]
    )
    def test_userconfig_set_valid(self, defaults, value, tmpdir):
        name = 'foobar'
        path = str(tmpdir)
        conf = UserConfig(name=name, path=path, defaults=defaults,
                          load=False, version='1.0.0', backup=False,
                          raw_mode=True)
        conf.set('test', 'opt', value)

    @pytest.mark.parametrize(
        'defaults,value',
        [
            ([('test', {'opt': 123})], 'no'),
            ([('test', {'opt': 123.123})], 'n9'),
            ([('test', {'opt': 123.123})], 'n9'),
        ]
    )
    def test_userconfig_set_invalid(self, defaults, value, tmpdir):
        name = 'foobar'
        path = str(tmpdir)
        conf = UserConfig(name=name, path=path, defaults=defaults,
                         load=False, version='1.0.0', backup=False,
                          raw_mode=True)

        with pytest.raises(ValueError):
            conf.set('test', 'opt', value)

    def test_userconfig_set_with_string(self, userconfig):
        userconfig.set('section', 'option', 'new value')
        with open(userconfig.get_config_fpath()) as inifile:
            ini_contents = inifile.read()

        expected = '[main]\nversion = 1.0.0\n\n'
        expected += "[section]\noption = new value\n\n"

        assert ini_contents == expected

    def test_userconfig_set_percentage_string(self, userconfig):
        """Test to set an option with a '%'."""
        userconfig.set('section', 'option', '%value')
        assert userconfig.get('section', 'option') == '%value'


def test_userconfig_remove_section(userconfig):
    assert 'section' in userconfig.sections()
    userconfig.remove_section('section')
    assert 'section' not in userconfig.sections()


def test_userconfig_remove_option(userconfig):
    assert userconfig.get('section', 'option') == 'value'
    userconfig.remove_option('section', 'option')
    with pytest.raises(cp.NoOptionError):
        userconfig.get('section', 'option')


def test_userconfig_cleanup(userconfig):
    configpath = userconfig.get_config_fpath()
    assert os.path.isfile(configpath)
    userconfig.cleanup()
    assert not os.path.isfile(configpath)


@pytest.mark.no_reset_conf
def test_invalid_shortcuts(tmp_path):
    name = 'invalid-shortcuts'
    path = str(tmp_path)

    for defaults in [
        # Shortcut is not of the form context/name
        [('shortcuts', {'foo': "Ctrl+I"})],
        # Shortcut contains more than one slash, which breaks the way we parse
        # them
        [('shortcuts', {'editor/foo/bar': "Ctrl+I"})],
    ]:
        with pytest.raises(ValueError):
            UserConfig(
                name=name,
                path=path,
                defaults=defaults,
                load=False,
                version="1.0.0",
                backup=False,
                raw_mode=True,
            )


# --- SpyderUserConfig tests
# ============================================================================
# --- Compatibility API
class TestSpyderConfigApplyPatches:

    def test_spyderconfig_apply_configuration_patches_42(
            self, spyderconfig_patches_42):
        # Check that the value is updated
        value = spyderconfig_patches_42.get('ipython_console',
                                            'startup/run_lines')
        expected_value = 'value1; value2'
        assert value == expected_value

    def test_spyderconfig_apply_configuration_patches_45(
            self, spyderconfig_patches_45):
        # Check that the value is not updated
        value = spyderconfig_patches_45.get('ipython_console',
                                            'startup/run_lines')
        expected_value = 'value1,value2'
        assert value == expected_value


def test_spyderconfig_get_defaults_path_name_from_version(spyderconfig):
    func = spyderconfig.get_defaults_path_name_from_version
    _, name = func('50.0.0')
    assert name == 'defaults-50.0.0'

    path, name = func('51.0.0')
    assert name == 'defaults-spyder-test-51.0.0'
    assert path.endswith('defaults')

    path, name = func('53.0.0')
    assert name == 'defaults-spyder-test-53.0.0'
    assert path.endswith('defaults')


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Language Server Protocol linting configuration tab.
"""

# Third party imports
from qtpy.QtWidgets import QLabel, QVBoxLayout

# Local imports
from spyder.api.preferences import SpyderPreferencesTab
from spyder.config.base import _


class LintingConfigTab(SpyderPreferencesTab):
    """Linting configuration tab."""

    TITLE = _('Linting')

    def __init__(self, parent):
        super().__init__(parent)
        newcb = self.create_checkbox

        linting_label = QLabel(_("Spyder can optionally highlight syntax "
                                 "errors and possible problems with your "
                                 "code in the editor."))
        linting_label.setOpenExternalLinks(True)
        linting_label.setWordWrap(True)
        linting_check = self.create_checkbox(
            _("Enable basic linting"),
            'pyflakes')
        underline_errors_box = newcb(
            _("Underline errors and warnings"),
            'underline_errors',
            section='editor')
        linting_complexity_box = self.create_checkbox(
            _("Enable complexity linting with the Mccabe package"),
            'mccabe')

        # Linting layout
        linting_layout = QVBoxLayout()
        linting_layout.addWidget(linting_label)
        linting_layout.addWidget(linting_check)
        linting_layout.addWidget(underline_errors_box)
        linting_layout.addWidget(linting_complexity_box)
        self.setLayout(linting_layout)
        linting_check.checkbox.toggled.connect(underline_errors_box.setEnabled)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Widgets to extend Spyder through its API.
"""


class PluginMainWidgetWidgets:
    CornerWidget = 'corner_widget'
    MainToolbar = 'main_toolbar_widget'
    OptionsToolButton = 'options_button_widget'
    Spinner = 'spinner_widget'


class PluginMainWidgetActions:
    ClosePane = 'close_pane'
    DockPane = 'dock_pane'
    UndockPane = 'undock_pane'
    LockUnlockPosition = 'lock_unlock_position'



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Debugger Plugin Configuration Page."""

# Third party imports
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QGroupBox, QVBoxLayout, QLabel

# Local imports
from spyder.config.base import _
from spyder.api.preferences import PluginConfigPage


class DebuggerConfigPage(PluginConfigPage):

    def setup_page(self):
        newcb = self.create_checkbox

        # ---- Debug ----
        # Pdb run lines Group
        pdb_run_lines_group = QGroupBox(_("Run code while debugging"))
        pdb_run_lines_label = QLabel(_(
            "You can run several lines of code on each new prompt while "
            "debugging. Please introduce each one separated by semicolons and "
            "a space, for example:<br>"
            "<i>import matplotlib.pyplot as plt; import numpy as np</i>")
        )
        pdb_run_lines_label.setWordWrap(True)
        pdb_run_lines_edit = self.create_lineedit(
            _("Lines:"), 'startup/pdb_run_lines', '', alignment=Qt.Horizontal)

        pdb_run_lines_layout = QVBoxLayout()
        pdb_run_lines_layout.addWidget(pdb_run_lines_label)
        pdb_run_lines_layout.addWidget(pdb_run_lines_edit)
        pdb_run_lines_group.setLayout(pdb_run_lines_layout)

        # Debug Group
        debug_group = QGroupBox(_("Interaction"))
        debug_layout = QVBoxLayout()

        prevent_closing_box = newcb(
            _("Prevent editor from closing files while debugging"),
            'pdb_prevent_closing',
            tip=_("This option prevents the user from closing a file while"
                  " it is debugged."))
        debug_layout.addWidget(prevent_closing_box)

        continue_box = newcb(
            _("Stop debugging on first line of files without breakpoints"),
            'pdb_stop_first_line',
            tip=_("This option lets you decide if the debugger should"
                  " stop on the first line while debugging if no breakpoints"
                  " are present."))
        debug_layout.addWidget(continue_box)

        libraries_box = newcb(
            _("Ignore Python libraries while debugging"), 'pdb_ignore_lib',
            tip=_("This option lets you decide if the debugger should "
                  "ignore the system libraries while debugging."))
        debug_layout.addWidget(libraries_box)

        execute_events_box = newcb(
            _("Process execute events while debugging"), 'pdb_execute_events',
            tip=_("This option lets you decide if the debugger should "
                  "process the 'execute events' after each prompt, such as "
                  "matplotlib <tt>show</tt> command."))
        debug_layout.addWidget(execute_events_box)

        exclamation_mark_box = newcb(
            _("Use exclamation mark prefix for Pdb commands"),
            'pdb_use_exclamation_mark',
            tip=_("This option lets you decide if the Pdb commands should "
                  "be prefixed by an exclamation mark. This helps in "
                  "separating Pdb commands from Python code."))
        debug_layout.addWidget(exclamation_mark_box)

        debug_group.setLayout(debug_layout)

        filter_group = QGroupBox(_("Execution Inspector"))
        filter_data = [
            ('exclude_internal', _("Exclude internal frames when inspecting execution")),
        ]
        filter_boxes = [self.create_checkbox(text, option)
                        for option, text in filter_data]

        filter_layout = QVBoxLayout()
        for box in filter_boxes:
            filter_layout.addWidget(box)
        filter_group.setLayout(filter_layout)

        vlayout = QVBoxLayout()
        vlayout.addWidget(debug_group)
        vlayout.addWidget(pdb_run_lines_group)
        vlayout.addWidget(filter_group)
        vlayout.addStretch(1)
        self.setLayout(vlayout)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright © Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------

"""
Tests for finding plugins.
"""

import pytest

from spyder.api.plugins import Plugins
from spyder.api.utils import get_class_values
from spyder.app.find_plugins import (
    find_internal_plugins, find_external_plugins)
from spyder.config.base import running_in_ci


def test_find_internal_plugins():
    """Test that we return all internal plugins available."""
    # We don't take the 'All' plugin into account here because it's not
    # really a plugin.
    expected_names = get_class_values(Plugins)
    expected_names.remove(Plugins.All)

    # Dictionary of internal plugins
    internal_plugins = find_internal_plugins()

    # Lengths must be the same
    assert len(expected_names) == len(internal_plugins.values())

    # Names must be the same
    assert sorted(expected_names) == sorted(list(internal_plugins.keys()))


@pytest.mark.skipif(not running_in_ci(), reason="Only works in CIs")
def test_find_external_plugins():
    """Test that we return the external plugins installed when testing."""
    internal_names = get_class_values(Plugins)
    expected_names = ['spyder_boilerplate']
    expected_special_attrs = {
        'spyder_boilerplate': [
            'spyder_boilerplate.spyder.plugin',
            'spyder-boilerplate',
            '0.0.1'
        ]
    }

    # Dictionary of external plugins
    external_plugins = find_external_plugins()

    # External plugins must be the ones installed while testing
    assert len(external_plugins.keys()) == len(expected_names)

    # Names must not be among internal plugins
    for name in external_plugins.keys():
        assert name not in internal_names

    # Names must be the expected ones.
    assert sorted(expected_names) == sorted(list(external_plugins.keys()))

    # Assert special attributes are present
    for name in external_plugins.keys():
        plugin_class = external_plugins[name]
        special_attrs = [
            plugin_class._spyder_module_name,
            plugin_class._spyder_package_name,
            plugin_class._spyder_version
        ]

        assert expected_special_attrs[name] == special_attrs



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""
Tests for pydocgui.py
"""
# Standard library imports
import os
import sys
from unittest.mock import MagicMock

# Test library imports
import numpy as np
from numpy.lib import NumpyVersion
import pytest
from flaky import flaky

# Local imports
from spyder.config.base import running_in_ci
from spyder.plugins.onlinehelp.widgets import PydocBrowser


@pytest.fixture
def pydocbrowser(qtbot):
    """Set up pydocbrowser."""
    plugin_mock = MagicMock()
    plugin_mock.CONF_SECTION = 'onlinehelp'
    widget = PydocBrowser(parent=None, plugin=plugin_mock, name='pydoc')
    with qtbot.waitSignal(widget.webview.loadFinished, timeout=20000):
        widget._setup()
        widget.setup()
        widget.resize(640, 480)
        widget.show()
        widget.initialize(force=True)
    yield widget
    widget.close()


@flaky(max_runs=5)
@pytest.mark.order(1)
@pytest.mark.parametrize(
    "lib",
    [('str', 'class str', [1, 2]),
     ('numpy.testing', 'numpy.testing', [5, 10]),
     ('numpy.finfo', 'numpy.finfo', [1, 5])]
)
def test_get_pydoc(pydocbrowser, qtbot, lib):
    """
    Go to the documentation by url.
    Regression test for spyder-ide/spyder#10740
    """
    browser = pydocbrowser
    element, doc, matches = lib

    webview = browser.webview
    element_url = browser.text_to_url(element)
    with qtbot.waitSignal(webview.loadFinished):
        browser.set_url(element_url)

    expected_range = list(range(matches[0], matches[1]))
    qtbot.waitUntil(lambda: webview.get_number_matches(doc) in expected_range)


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

import uuid
import difflib
import logging
import textwrap

from spyder.plugins.remoteclient.api.jupyterhub import JupyterHubAPI
from spyder.plugins.remoteclient.api.jupyterhub.utils import (
    parse_notebook_cells,
)

logger = logging.getLogger(__name__)


DAEMONIZED_STOP_SERVER_HEADER = """
def _client_stop_server():
    import urllib.request
    request = urllib.request.Request(url="{delete_server_endpoint}", method= "DELETE")
    request.add_header("Authorization", "token {api_token}")
    urllib.request.urlopen(request)

def custom_exc(shell, etype, evalue, tb, tb_offset=None):
     _jupyerhub_client_stop_server()

get_ipython().set_custom_exc((Exception,), custom_exc)
"""


async def determine_username(
    hub,
    username=None,
    user_format="user-{user}-{id}",
    service_format="service-{name}-{id}",
    temporary_user=False,
):
    token = await hub.identify_token(hub.api_token)

    if username is None and not temporary_user:
        if token["kind"] == "service":
            logger.error(
                "cannot execute without specified username or "
                "temporary_user=True for service api token"
            )
            raise ValueError(
                "Service api token cannot execute without specified username "
                "or temporary_user=True for"
            )
        return token["name"]
    elif username is None and temporary_user:
        if token["kind"] == "service":
            return service_format.format(
                id=str(uuid.uuid4()), name=token["name"]
            )
        else:
            return user_format.format(id=str(uuid.uuid4()), name=token["name"])
    else:
        return username


async def execute_code(
    hub_url,
    cells,
    username=None,
    temporary_user=False,
    create_user=False,
    delete_user=False,
    server_creation_timeout=60,
    server_deletion_timeout=60,
    kernel_execution_timeout=60,
    daemonized=False,
    validate=False,
    stop_server=True,
    user_options=None,
    kernel_spec=None,
    auth_type="token",
    verify_ssl=True,
):
    hub = JupyterHubAPI(hub_url, auth_type=auth_type, verify_ssl=verify_ssl)
    result_cells = []

    async with hub:
        username = await determine_username(
            hub, username, temporary_user=temporary_user
        )
        try:
            jupyter = await hub.ensure_server(
                username,
                create_user=create_user,
                user_options=user_options,
                timeout=server_creation_timeout,
            )

            async with jupyter:
                kernel_id, kernel = await jupyter.ensure_kernel(
                    kernel_spec=kernel_spec
                )
                async with kernel:
                    if daemonized and stop_server:
                        await kernel.send_code(
                            username,
                            DAEMONIZED_STOP_SERVER_HEADER.format(
                                delete_server_endpoint=hub.api_url
                                / "users"
                                / username
                                / "server",
                                api_token=hub.api_token,
                            ),
                            wait=False,
                        )

                    for i, (code, expected_result) in enumerate(cells):
                        kernel_result = await kernel.send_code(
                            username,
                            code,
                            timeout=kernel_execution_timeout,
                            wait=(not daemonized),
                        )
                        result_cells.append((code, kernel_result))
                        if daemonized:
                            logger.debug(
                                f"kernel submitted cell={i} "
                                f'code=\n{textwrap.indent(code, "   >>> ")}'
                            )
                        else:
                            logger.debug(
                                f"kernel executing cell={i} "
                                f'code=\n{textwrap.indent(code, "   >>> ")}'
                            )
                            logger.debug(
                                f"kernel result cell={i} result=\n"
                                f'{textwrap.indent(kernel_result, "   | ")}'
                            )
                            if validate and (
                                kernel_result.strip()
                                != expected_result.strip()
                            ):
                                diff = "".join(
                                    difflib.unified_diff(
                                        kernel_result, expected_result
                                    )
                                )
                                logger.error(
                                    f"kernel result did not match expected "
                                    f"result diff={diff}"
                                )
                                raise ValueError(
                                    f"execution of cell={i} did not match "
                                    f"expected result diff={diff}"
                                )

                    if daemonized and stop_server:
                        await kernel.send_code(
                            username, "__client_stop_server()", wait=False
                        )
                if not daemonized:
                    await jupyter.delete_kernel(kernel_id)
            if not daemonized and stop_server:
                await hub.ensure_server_deleted(
                    username, timeout=server_deletion_timeout
                )
        finally:
            if delete_user and not daemonized:
                await hub.delete_user(username)

        return result_cells


async def execute_notebook(hub_url, notebook_path, **kwargs):
    cells = parse_notebook_cells(notebook_path)
    return await execute_code(hub_url, cells, **kwargs)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""
Tests for shortcuts.py
"""

# Standard library imports
import os
import sys

# Third party imports
import pytest
from qtpy.QtCore import Qt

# Local imports
from spyder.config.base import running_in_ci
from spyder.config.manager import CONF
from spyder.plugins.shortcuts.utils import ShortcutData
from spyder.plugins.shortcuts.widgets.table import (
    INVALID_KEY, NO_WARNING, SEQUENCE_CONFLICT, SEQUENCE_EMPTY,
    ShortcutEditor, ShortcutsTable, load_shortcuts)


# ---- Qt Test Fixtures
@pytest.fixture
def shortcut_table(qtbot):
    """Set up shortcuts."""
    shortcut_table = ShortcutsTable()
    shortcut_table = load_shortcuts(shortcut_table)
    qtbot.addWidget(shortcut_table)
    return shortcut_table


@pytest.fixture
def create_shortcut_editor(shortcut_table, qtbot):
    shortcuts = shortcut_table.source_model.shortcuts

    def _create_bot(context, name):
        sequence = CONF.get('shortcuts', "{}/{}".format(context, name))
        shortcut_editor = ShortcutEditor(
            shortcut_table, context, name, sequence, shortcuts)
        qtbot.addWidget(shortcut_editor)
        shortcut_editor.show()
        return shortcut_editor
    return _create_bot


# ---- Filter text mock
class FilterTextMock():
    def __init__(self, text):
        self.txt = text

    def text(self):
        return self.txt


# ---- Tests ShortcutsTable
@pytest.mark.skipif(
    sys.platform.startswith('linux') and running_in_ci(),
    reason="It fails on Linux due to the lack of a proper X server.")
def test_shortcuts(shortcut_table):
    """Run shortcuts table."""
    shortcut_table.show()
    shortcut_table.check_shortcuts()
    assert shortcut_table


def test_shortcut_in_conf_is_filtered_with_shortcut_data(qtbot):
    shortcut_table = ShortcutsTable()
    shortcut_table = load_shortcuts(shortcut_table)
    qtbot.addWidget(shortcut_table)
    row_count = shortcut_table.model().rowCount()
    assert row_count != 0

    shortcut_table_empty = ShortcutsTable()
    shortcut_table_empty.set_shortcut_data([
        ShortcutData(qobject=None, name='switch to plots', context='_'),
        ShortcutData(qobject=None, name='switch to editor', context='_')
    ])
    shortcut_table_empty.load_shortcuts()
    qtbot.addWidget(shortcut_table_empty)
    row_count = shortcut_table_empty.model().rowCount()
    assert row_count == 2


def test_shortcuts_filtering(shortcut_table):
    """Run shortcuts table."""
    # Store original row count
    row_count = shortcut_table.model().rowCount()
    # Filter for "debug"
    shortcut_table.finder = FilterTextMock('debug')
    shortcut_table.set_regex()
    # Sorting should be disabled
    assert not shortcut_table.isSortingEnabled()
    # Six hits (causes a bit of an issue to hardcode it like this if new
    # shortcuts are added...)
    assert shortcut_table.model().rowCount() == 14
    # Remove filter text
    shortcut_table.finder = FilterTextMock('')
    shortcut_table.set_regex()
    # Should be sortable again
    assert shortcut_table.isSortingEnabled()
    # All entries restored
    assert shortcut_table.model().rowCount() == row_count

    # Same thing, but using reset instead
    shortcut_table.finder = FilterTextMock('debug')
    shortcut_table.set_regex()
    shortcut_table.set_regex(reset=True)
    assert shortcut_table.isSortingEnabled()
    assert shortcut_table.model().rowCount() == row_count


def test_shortcut_filtering_context(shortcut_table):
    """Test multifiltering by context and name in the table."""
    # Verify that the model is filtering by two columns
    assert len(shortcut_table.model().filters) == 2
    # Filter by "console"
    shortcut_table.finder = FilterTextMock('console')
    shortcut_table.set_regex()
    # Verify the number of entries after the regex are 12
    # If a new shortcut is added to console, this needs to be changed
    assert shortcut_table.model().rowCount() == 12

    # Filter by "pylint"
    shortcut_table.finder = FilterTextMock('pylint')
    shortcut_table.set_regex()
    # Verify the number of entries after the regex is 1
    # If a new shortcut is added to pylint, this needs to be changed
    assert shortcut_table.model().rowCount() == 2


# ---- Tests ShortcutEditor
def test_clear_shortcut(create_shortcut_editor, qtbot):
    """
    Test that pressing on the 'Clear' button to unbind the command from a
    shortcut is working as expected.
    """
    shortcut_editor = create_shortcut_editor('editor', 'delete line')
    qtbot.mouseClick(shortcut_editor.button_clear, Qt.LeftButton)
    assert shortcut_editor.new_sequence == ''


def test_press_new_sequence(create_shortcut_editor, qtbot):
    """
    Test that pressing a key sequence with modifier keys is registered as
    expected by the Shortcut Editor.
    """
    shortcut_editor = create_shortcut_editor('editor', 'delete line')
    modifiers = Qt.ControlModifier | Qt.ShiftModifier | Qt.AltModifier
    qtbot.keyClick(shortcut_editor, Qt.Key_D, modifier=modifiers)
    assert shortcut_editor.new_sequence == 'Ctrl+Alt+Shift+D'
    assert shortcut_editor.warning == NO_WARNING
    assert shortcut_editor.button_ok.isEnabled()


def test_press_new_compound_sequence(create_shortcut_editor, qtbot):
    """
    Test that pressing a compund of key sequences is registered as
    expected by the Shortcut Editor.
    """
    shortcut_editor = create_shortcut_editor('editor', 'delete line')
    qtbot.keyClick(shortcut_editor, Qt.Key_D, modifier=Qt.ControlModifier)
    qtbot.keyClick(shortcut_editor, Qt.Key_A)
    qtbot.keyClick(shortcut_editor, Qt.Key_B, modifier=Qt.ControlModifier)
    qtbot.keyClick(shortcut_editor, Qt.Key_C)
    qtbot.keyClick(shortcut_editor, Qt.Key_D)
    assert shortcut_editor.new_sequence == 'Ctrl+D, A, Ctrl+B, C'
    # The 'D' key press event is discarted because a compound sequence
    # cannot be composed of more than 4 sub sequences.
    assert shortcut_editor.warning == NO_WARNING
    assert shortcut_editor.button_ok.isEnabled()


def test_clear_back_new_sequence(create_shortcut_editor, qtbot):
    """
    Test that removing the last key sequence entered and clearing all entered
    key sequence from the Shortcut Editor is working as expected.
    """
    shortcut_editor = create_shortcut_editor('editor', 'delete line')
    qtbot.keyClick(shortcut_editor, Qt.Key_X, modifier=Qt.ControlModifier)
    qtbot.keyClick(shortcut_editor, Qt.Key_A)
    qtbot.keyClick(shortcut_editor, Qt.Key_B, modifier=Qt.ControlModifier)
    qtbot.keyClick(shortcut_editor, Qt.Key_C)
    qtbot.keyClick(shortcut_editor, Qt.Key_D)

    # Remove last key sequence entered.
    qtbot.mouseClick(shortcut_editor.button_back_sequence, Qt.LeftButton)
    assert shortcut_editor.new_sequence == 'Ctrl+X, A, Ctrl+B'
    assert shortcut_editor.warning == SEQUENCE_CONFLICT
    assert shortcut_editor.button_ok.isEnabled()

    # Remove second to last key sequence entered.
    qtbot.mouseClick(shortcut_editor.button_back_sequence, Qt.LeftButton)
    assert shortcut_editor.new_sequence == 'Ctrl+X, A'
    assert shortcut_editor.warning == SEQUENCE_CONFLICT
    assert shortcut_editor.button_ok.isEnabled()

    # Clear all entered key sequences.
    qtbot.mouseClick(shortcut_editor.btn_clear_sequence, Qt.LeftButton)
    assert shortcut_editor.new_sequence == ''
    assert shortcut_editor.warning == SEQUENCE_EMPTY
    assert not shortcut_editor.button_ok.isEnabled()


def test_sequence_conflict(create_shortcut_editor, qtbot):
    """
    Test that the Shortcut Editor is able to detect key sequence conflict
    with other shortcuts.
    """
    shortcut_editor = create_shortcut_editor('editor', 'delete line')

    # Check that the conflict is detected for a single key sequence.
    qtbot.keyClick(shortcut_editor, Qt.Key_X, modifier=Qt.ControlModifier)
    assert shortcut_editor.new_sequence == 'Ctrl+X'
    assert shortcut_editor.warning == SEQUENCE_CONFLICT
    assert shortcut_editor.button_ok.isEnabled()


    # Check that the conflict is detected for a compound of key sequences.
    qtbot.keyClick(shortcut_editor, Qt.Key_X)
    assert shortcut_editor.new_sequence == 'Ctrl+X, X'
    assert shortcut_editor.warning == SEQUENCE_CONFLICT
    assert shortcut_editor.button_ok.isEnabled()


def test_sequence_single_key(create_shortcut_editor, qtbot):
    """
    Test that the Shortcut Editor raise a warning when the first key
    sequence entered is composed of a single key with no modifier and this
    single key is not in the list of supported single key sequence.
    """
    shortcut_editor = create_shortcut_editor('editor', 'delete line')

    # Check this is working as expected for a single key sequence.
    qtbot.keyClick(shortcut_editor, Qt.Key_D)
    assert shortcut_editor.new_sequence == 'D'
    assert shortcut_editor.warning == INVALID_KEY
    assert not shortcut_editor.button_ok.isEnabled()

    # Check this is working as expected for a compound of key sequences.
    qtbot.keyClick(shortcut_editor, Qt.Key_D, modifier=Qt.ControlModifier)
    assert shortcut_editor.new_sequence == 'D, Ctrl+D'
    assert shortcut_editor.warning == INVALID_KEY
    assert not shortcut_editor.button_ok.isEnabled()

    # Check this is working as expected when a valid single key is pressed.
    qtbot.mouseClick(shortcut_editor.btn_clear_sequence, Qt.LeftButton)
    qtbot.keyClick(shortcut_editor, Qt.Key_Home)
    assert shortcut_editor.new_sequence == 'Home'
    assert shortcut_editor.warning == NO_WARNING
    assert shortcut_editor.button_ok.isEnabled()


def test_set_sequence_to_default(create_shortcut_editor, qtbot):
    """
    Test that clicking on the button 'Default' set the sequence in the
    Shortcut Editor to the default value as espected.
    """
    shortcut_editor = create_shortcut_editor('editor', 'delete line')
    default_sequence = CONF.get(
        'shortcuts', "{}/{}".format('editor', 'delete line'))

    qtbot.mouseClick(shortcut_editor.button_default, Qt.LeftButton)
    assert shortcut_editor.new_sequence == default_sequence
    assert shortcut_editor.warning == NO_WARNING
    assert shortcut_editor.button_ok.isEnabled()


def test_invalid_char_in_sequence(create_shortcut_editor, qtbot):
    """
    Test that the key sequence is rejected and a warning is shown if an
    invalid character is present in the new key sequence.
    """
    shortcut_editor = create_shortcut_editor('editor', 'delete line')

    # Check this is working as expected for a single key sequence.
    qtbot.keyClick(shortcut_editor, Qt.Key_Odiaeresis,
                   modifier=Qt.ControlModifier | Qt.AltModifier)
    assert shortcut_editor.warning == INVALID_KEY
    assert not shortcut_editor.button_ok.isEnabled()


if __name__ == "__main__":
    pytest.main(['-x', os.path.basename(__file__), '-vv', '-rw'])



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Scroll flag panel for the editor.
"""

# Standard library imports
import logging
from math import ceil
import sys

# Third party imports
from qtpy.QtCore import QSize, Qt, QThread
from qtpy.QtGui import QColor, QCursor, QPainter
from qtpy.QtWidgets import QApplication, QStyle, QStyleOptionSlider
from superqt.utils import qdebounced

# Local imports
from spyder.plugins.completion.api import DiagnosticSeverity
from spyder.plugins.editor.api.panel import Panel
from spyder.plugins.editor.utils.editor import is_block_safe


# For logging
logger = logging.getLogger(__name__)

# Time to wait before refreshing flags
REFRESH_RATE = 1000

# Maximum number of flags to paint in a file
MAX_FLAGS = 1000


class ScrollFlagArea(Panel):
    """Source code editor's scroll flag area"""
    WIDTH = 24 if sys.platform == 'darwin' else 12
    FLAGS_DX = 4
    FLAGS_DY = 2

    def __init__(self):
        Panel.__init__(self)
        self.setAttribute(Qt.WA_OpaquePaintEvent)
        self.scrollable = True
        self.setMouseTracking(True)

        # Define some attributes to be used for unit testing.
        self._unit_testing = False
        self._range_indicator_is_visible = False
        self._alt_key_is_down = False

        self._slider_range_color = QColor(Qt.gray)
        self._slider_range_color.setAlphaF(.85)
        self._slider_range_brush = QColor(Qt.gray)
        self._slider_range_brush.setAlphaF(.5)

        # Dictionary with flag lists
        self._dict_flag_list = {}

        # Thread to update flags on it.
        self._update_flags_thread = QThread(None)
        self._update_flags_thread.run = self._update_flags
        self._update_flags_thread.finished.connect(self.update)

    def on_install(self, editor):
        """Manages install setup of the pane."""
        super().on_install(editor)
        # Define permanent Qt colors that are needed for painting the flags
        # and the slider range.
        self._facecolors = {
            'warning': QColor(editor.warning_color),
            'error': QColor(editor.error_color),
            'todo': QColor(editor.todo_color),
            'breakpoint': QColor(editor.breakpoint_color),
            'occurrence': QColor(editor.occurrence_color),
            'found_results': QColor(editor.found_results_color)
            }
        self._edgecolors = {key: color.darker(120) for
                            key, color in self._facecolors.items()}

        # Signals
        editor.sig_focus_changed.connect(self.update)
        editor.sig_key_pressed.connect(self.keyPressEvent)
        editor.sig_key_released.connect(self.keyReleaseEvent)
        editor.sig_alt_left_mouse_pressed.connect(self.mousePressEvent)
        editor.sig_alt_mouse_moved.connect(self.mouseMoveEvent)
        editor.sig_leave_out.connect(self.update)
        editor.sig_flags_changed.connect(self.update_flags)
        editor.sig_theme_colors_changed.connect(self.update_flag_colors)

        # This prevents that flags are updated while the user is moving the
        # cursor, e.g. when typing.
        editor.sig_cursor_position_changed.connect(self.update_flags)

    @property
    def slider(self):
        """This property holds whether the vertical scrollbar is visible."""
        return self.editor.verticalScrollBar().isVisible()

    def closeEvent(self, event):
        self._update_flags_thread.quit()
        self._update_flags_thread.wait()
        super().closeEvent(event)

    def sizeHint(self):
        """Override Qt method"""
        return QSize(self.WIDTH, 0)

    def update_flag_colors(self, color_dict):
        """
        Update the permanent Qt colors that are used for painting the flags
        and the slider range with the new colors defined in the given dict.
        """
        for name, color in color_dict.items():
            self._facecolors[name] = QColor(color)
            self._edgecolors[name] = self._facecolors[name].darker(120)

    @qdebounced(timeout=REFRESH_RATE)
    def update_flags(self):
        """Update flags list in a thread."""
        logger.debug("Updating current flags")

        self._dict_flag_list = {
            'error': [],
            'warning': [],
            'todo': [],
            'breakpoint': [],
        }

        # Run this computation in a different thread to prevent freezing 
        # the interface
        if not self._update_flags_thread.isRunning():
            self._update_flags_thread.start()

    def _update_flags(self):
        """Update flags list."""
        editor = self.editor
        block = editor.document().firstBlock()
        while block.isValid():
            # Parse all lines in the file looking for something to flag.
            data = block.userData()
            if data:
                if data.code_analysis:
                    for _, _, severity, _ in data.code_analysis:
                        if severity == DiagnosticSeverity.ERROR:
                            flag_type = 'error'
                            break
                    else:
                        flag_type = 'warning'
                elif data.todo:
                    flag_type = 'todo'
                elif data.breakpoint:
                    flag_type = 'breakpoint'
                else:
                    flag_type = None

                if flag_type is not None:
                    self._dict_flag_list[flag_type].append(block)

            block = block.next()

    def paintEvent(self, event):
        """
        Override Qt method.
        Painting the scroll flag area

        There is two cases:
            - The scroll bar is moving, in which case paint all flags.
            - The scroll bar is not moving, only paint flags corresponding
              to visible lines.
        """
        # The area in which the slider handle of the scrollbar may move.
        groove_rect = self.get_scrollbar_groove_rect()

        # This is necessary to catch a possible error when the scrollbar
        # has zero height.
        # Fixes spyder-ide/spyder#21600
        try:
            # The scrollbar's scale factor ratio between pixel span height and
            # value span height
            scale_factor = (
                groove_rect.height() / self.get_scrollbar_value_height()
            )
        except ZeroDivisionError:
            scale_factor = 1

        # The vertical offset of the scroll flag area relative to the
        # top of the text editor.
        offset = groove_rect.y()

        # Note that we calculate the pixel metrics required to draw the flags
        # here instead of using the convenience methods of the ScrollFlagArea
        # for performance reason.
        rect_x = ceil(self.FLAGS_DX / 2)
        rect_w = self.WIDTH - self.FLAGS_DX
        rect_h = self.FLAGS_DY

        # Fill the whole painting area
        painter = QPainter(self)
        painter.fillRect(event.rect(), self.editor.sideareas_color)

        editor = self.editor

        # Define compute_flag_ypos to position the flags:
        # Paint flags for the entire document
        last_line = editor.document().lastBlock().firstLineNumber()
        # The 0.5 offset is used to align the flags with the center of
        # their corresponding text edit block before scaling.
        first_y_pos = self.value_to_position(
            0.5, scale_factor, offset) - self.FLAGS_DY / 2
        last_y_pos = self.value_to_position(
            last_line + 0.5, scale_factor, offset) - self.FLAGS_DY / 2

        # Compute the height of a line and of a flag in lines.
        line_height = last_y_pos - first_y_pos
        if line_height > 0:
            flag_height_lines = rect_h * last_line / line_height
        else:
            flag_height_lines = 0

        # All the lists of block numbers for flags
        dict_flag_lists = {
            "occurrence": editor.occurrences,
            "found_results": editor.found_results
        }
        dict_flag_lists.update(self._dict_flag_list)

        # The ability to reverse dictionaries was added in Python 3.8.
        # Fixes spyder-ide/spyder#21286
        if sys.version_info[:2] > (3, 7):
            # This is necessary to paint find matches above errors and
            # warnings.
            # See spyder-ide/spyder#20970
            dict_flag_lists_iter = reversed(dict_flag_lists)
        else:
            dict_flag_lists_iter = dict_flag_lists

        for flag_type in dict_flag_lists_iter:
            painter.setBrush(self._facecolors[flag_type])
            painter.setPen(self._edgecolors[flag_type])
            if editor.verticalScrollBar().maximum() == 0:
                # No scroll
                for block in dict_flag_lists[flag_type]:
                    if not is_block_safe(block):
                        continue
                    geometry = editor.blockBoundingGeometry(block)
                    rect_y = ceil(
                        geometry.y() +
                        geometry.height() / 2 +
                        rect_h / 2
                    )
                    painter.drawRect(rect_x, rect_y, rect_w, rect_h)
            elif last_line == 0:
                # Only one line
                for block in dict_flag_lists[flag_type]:
                    if not is_block_safe(block):
                        continue
                    rect_y = ceil(first_y_pos)
                    painter.drawRect(rect_x, rect_y, rect_w, rect_h)
            else:
                # Many lines
                if len(dict_flag_lists[flag_type]) < MAX_FLAGS:
                    # If the file is too long, do not freeze the editor
                    next_line = 0
                    for block in dict_flag_lists[flag_type]:
                        if not is_block_safe(block):
                            continue
                        block_line = block.firstLineNumber()
                        # block_line = -1 if invalid
                        if block_line < next_line:
                            # Don't print flags on top of flags
                            continue
                        next_line = block_line + flag_height_lines / 2
                        frac = block_line / last_line
                        rect_y = ceil(first_y_pos + frac * line_height)
                        painter.drawRect(rect_x, rect_y, rect_w, rect_h)

        # Paint the slider range
        if not self._unit_testing:
            alt = QApplication.queryKeyboardModifiers() & Qt.AltModifier
        else:
            alt = self._alt_key_is_down

        if self.slider:
            cursor_pos = self.mapFromGlobal(QCursor().pos())
            is_over_self = self.rect().contains(cursor_pos)
            is_over_editor = editor.rect().contains(
                editor.mapFromGlobal(QCursor().pos()))
            # We use QRect.contains instead of QWidget.underMouse method to
            # determined if the cursor is over the editor or the flag scrollbar
            # because the later gives a wrong result when a mouse button
            # is pressed.
            if is_over_self or (alt and is_over_editor):
                painter.setPen(self._slider_range_color)
                painter.setBrush(self._slider_range_brush)
                x, y, width, height = self.make_slider_range(
                    cursor_pos, scale_factor, offset, groove_rect)
                painter.drawRect(x, y, width, height)
                self._range_indicator_is_visible = True
            else:
                self._range_indicator_is_visible = False

    def enterEvent(self, event):
        """Override Qt method"""
        self.update()

    def leaveEvent(self, event):
        """Override Qt method"""
        self.update()

    def mouseMoveEvent(self, event):
        """Override Qt method"""
        self.update()

    def mousePressEvent(self, event):
        """Override Qt method"""
        if self.slider and event.button() == Qt.LeftButton:
            vsb = self.editor.verticalScrollBar()
            value = self.position_to_value(event.pos().y())
            vsb.setValue(int(value-vsb.pageStep()/2))

    def keyReleaseEvent(self, event):
        """Override Qt method."""
        if event.key() == Qt.Key_Alt:
            self._alt_key_is_down = False
            self.update()

    def keyPressEvent(self, event):
        """Override Qt method"""
        if event.key() == Qt.Key_Alt:
            self._alt_key_is_down = True
            self.update()

    def get_vertical_offset(self):
        """
        Return the vertical offset of the scroll flag area relative to the
        top of the text editor.
        """
        groove_rect = self.get_scrollbar_groove_rect()
        return groove_rect.y()

    def get_slider_min_height(self):
        """
        Return the minimum height of the slider range based on that set for
        the scroll bar's slider.
        """
        return QApplication.instance().style().pixelMetric(
            QStyle.PM_ScrollBarSliderMin)

    def get_scrollbar_groove_rect(self):
        """Return the area in which the slider handle may move."""
        vsb = self.editor.verticalScrollBar()
        style = QApplication.instance().style()
        opt = QStyleOptionSlider()
        vsb.initStyleOption(opt)

        # Get the area in which the slider handle may move.
        groove_rect = style.subControlRect(
            QStyle.CC_ScrollBar, opt, QStyle.SC_ScrollBarGroove, self)

        return groove_rect

    def get_scrollbar_position_height(self):
        """Return the pixel span height of the scrollbar area in which
        the slider handle may move"""
        groove_rect = self.get_scrollbar_groove_rect()
        return float(groove_rect.height())

    def get_scrollbar_value_height(self):
        """Return the value span height of the scrollbar"""
        vsb = self.editor.verticalScrollBar()
        return vsb.maximum() - vsb.minimum() + vsb.pageStep()

    def get_scale_factor(self):
        """Return scrollbar's scale factor:
        ratio between pixel span height and value span height"""
        return (self.get_scrollbar_position_height() /
                self.get_scrollbar_value_height())

    def value_to_position(self, y, scale_factor, offset):
        """Convert value to position in pixels"""
        vsb = self.editor.verticalScrollBar()
        return int((y - vsb.minimum()) * scale_factor + offset)

    def position_to_value(self, y):
        """Convert position in pixels to value"""
        vsb = self.editor.verticalScrollBar()
        offset = self.get_vertical_offset()
        return vsb.minimum() + max([0, (y - offset) / self.get_scale_factor()])

    def make_slider_range(self, cursor_pos, scale_factor, offset, groove_rect):
        """
        Return the slider x and y positions and the slider width and height.
        """
        # The slider range indicator position follows the mouse vertical
        # position while its height corresponds to the part of the file that
        # is currently visible on screen.

        vsb = self.editor.verticalScrollBar()
        slider_height = self.value_to_position(
            vsb.pageStep(), scale_factor, offset) - offset
        slider_height = max(slider_height, self.get_slider_min_height())

        # Calculate the minimum and maximum y-value to constraint the slider
        # range indicator position to the height span of the scrollbar area
        # where the slider may move.
        min_ypos = offset
        max_ypos = groove_rect.height() + offset - slider_height

        # Determine the bounded y-position of the slider rect.
        slider_y = max(min_ypos, min(max_ypos,
                                     ceil(cursor_pos.y()-slider_height/2)))

        return 1, slider_y, self.WIDTH - 2, slider_height

    def wheelEvent(self, event):
        """Override Qt method"""
        self.editor.wheelEvent(event)

    def set_enabled(self, state):
        """Toggle scroll flag area visibility"""
        self.enabled = state
        self.setVisible(state)




# -*- coding: utf-8 -*-

# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Fallback completion actor.

This takes a plain text/source file and returns the individual words
written on it and the keywords associated by Pygments to the
programming language of that file.
"""

# Standard library imports
import logging

# Qt imports
from qtpy.QtCore import QObject, QThread, QMutex, QMutexLocker, Signal, Slot

# Other imports
from pygments.lexers import get_lexer_by_name
from diff_match_patch import diff_match_patch

# Local imports
from spyder.plugins.completion.api import CompletionItemKind
from spyder.plugins.completion.api import CompletionRequestTypes
from spyder.plugins.completion.providers.fallback.utils import (
    get_keywords, get_words, is_prefix_valid)


FALLBACK_COMPLETION = "Fallback"

logger = logging.getLogger(__name__)


class FallbackActor(QObject):
    #: Signal emitted when the Thread is ready
    sig_fallback_ready = Signal()
    sig_set_tokens = Signal(int, dict)
    sig_mailbox = Signal(dict)

    def __init__(self, parent):
        QObject.__init__(self)
        self.stopped = False
        self.daemon = True
        self.mutex = QMutex()
        self.file_tokens = {}
        self.diff_patch = diff_match_patch()
        self.thread = QThread(None)
        self.moveToThread(self.thread)

        self.thread.started.connect(self.started)
        self.sig_mailbox.connect(self.handle_msg)

    def tokenize(self, text, offset, language, current_word):
        """
        Return all tokens in `text` and all keywords associated by
        Pygments to `language`.
        """
        valid = is_prefix_valid(text, offset, language)
        if not valid:
            return []

        # Get language keywords provided by Pygments
        try:
            lexer = get_lexer_by_name(language)
            keywords = get_keywords(lexer)
        except Exception:
            keywords = []
        keyword_set = set(keywords)
        keywords = [{'kind': CompletionItemKind.KEYWORD,
                     'insertText': keyword,
                     'label': keyword,
                     'sortText': keyword,
                     'filterText': keyword,
                     'documentation': '',
                     'provider': FALLBACK_COMPLETION}
                    for keyword in keywords]

        # Get file tokens
        tokens = get_words(text, offset, language)
        tokens = [{'kind': CompletionItemKind.TEXT,
                   'insertText': token,
                   'label': token,
                   'sortText': token,
                   'filterText': token,
                   'documentation': '',
                   'provider': FALLBACK_COMPLETION}
                  for token in tokens]
        for token in tokens:
            if token['insertText'] not in keyword_set:
                keywords.append(token)

        # Filter matching results
        if current_word is not None:
            current_word = current_word.lower()
            keywords = [k for k in keywords
                        if current_word in k['insertText'].lower()]

        return keywords

    def stop(self):
        """Stop actor."""
        with QMutexLocker(self.mutex):
            logger.debug("Fallback plugin stopping...")
            self.thread.quit()
            self.thread.wait()

    def start(self):
        """Start thread."""
        self.thread.start()

    def started(self):
        """Thread started."""
        logger.debug('Fallback plugin starting...')
        self.sig_fallback_ready.emit()

    @Slot(dict)
    def handle_msg(self, message):
        """Handle one message"""
        msg_type, _id, file, msg = [
            message[k] for k in ('type', 'id', 'file', 'msg')]
        logger.debug(u'Perform request {0} with id {1}'.format(msg_type, _id))
        if msg_type == CompletionRequestTypes.DOCUMENT_DID_OPEN:
            self.file_tokens[file] = {
                'text': msg['text'],
                'offset': msg['offset'],
                'language': msg['language'],
            }
        elif msg_type == CompletionRequestTypes.DOCUMENT_DID_CHANGE:
            if file not in self.file_tokens:
                self.file_tokens[file] = {
                    'text': '',
                    'offset': msg['offset'],
                    'language': msg['language'],
                }
            diff = msg['diff']
            text = self.file_tokens[file]
            text['offset'] = msg['offset']
            text, _ = self.diff_patch.patch_apply(
                diff, text['text'])
            self.file_tokens[file]['text'] = text
        elif msg_type == CompletionRequestTypes.DOCUMENT_DID_CLOSE:
            self.file_tokens.pop(file, {})
        elif msg_type == CompletionRequestTypes.DOCUMENT_COMPLETION:
            tokens = []
            if file in self.file_tokens:
                text_info = self.file_tokens[file]
                tokens = self.tokenize(
                    text_info['text'],
                    text_info['offset'],
                    text_info['language'],
                    msg['current_word'])
            tokens = {'params': tokens}
            self.sig_set_tokens.emit(_id, tokens)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""Tests some cases were completions need to be hidden."""

# Third party imports
from flaky import flaky
import pytest
from qtpy.QtCore import Qt


@pytest.mark.order(1)
@flaky(max_runs=10)
def test_automatic_completions_hide_complete(completions_codeeditor, qtbot):
    """Test on-the-fly completion closing when already complete.

    Regression test for issue #11600 and pull requests #11824, #12140
    and #12710.
    """
    code_editor, _ = completions_codeeditor
    completion = code_editor.completion_widget
    delay = 50
    code_editor.toggle_code_snippets(False)

    code_editor.set_text('some = 0\nsomething = 1\n')
    cursor = code_editor.textCursor()
    code_editor.moveCursor(cursor.End)

    # Complete some -> [some, something]
    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyClicks(code_editor, 'some', delay=delay)
    assert "some" in [x['label'] for x in sig.args[0]]
    assert "something" in [x['label'] for x in sig.args[0]]

    # No completion for 'something' as already complete
    qtbot.keyClicks(code_editor, 'thing', delay=delay)
    qtbot.wait(500)
    assert completion.isHidden()
    qtbot.keyPress(code_editor, Qt.Key_Enter, delay=300)  # newline

    # Hide even within a function
    qtbot.keyClicks(code_editor, 'print(something', delay=delay)
    qtbot.wait(500)
    assert completion.isHidden()
    qtbot.keyClicks(code_editor, ')', delay=delay)
    qtbot.keyPress(code_editor, Qt.Key_Enter, delay=300)  # newline

    # Hide even inside comprehension
    qtbot.keyClicks(code_editor, 'a = {something', delay=delay)
    qtbot.wait(500)
    assert completion.isHidden()

    # Hide if removing spaces before a word
    code_editor.moveCursor(cursor.End)
    qtbot.keyPress(code_editor, Qt.Key_Enter)  # newline
    qtbot.keyClicks(code_editor, 'some', delay=delay)
    qtbot.keyPress(code_editor, Qt.Key_Enter)  # newline
    qtbot.keyClicks(code_editor, '  None', delay=delay)
    if completion.isVisible():
        qtbot.keyPress(completion, Qt.Key_Enter)
    code_editor.moveCursor(cursor.StartOfWord)
    qtbot.keyPress(code_editor, Qt.Key_Backspace)
    qtbot.wait(2000)
    assert completion.isHidden()
    qtbot.keyPress(code_editor, Qt.Key_Backspace)
    qtbot.wait(2000)
    assert completion.isHidden()

    # Hide if removing spaces before a word even not at the start of line.
    code_editor.moveCursor(cursor.End)
    qtbot.keyPress(code_editor, Qt.Key_Enter)  # newline
    qtbot.keyClicks(code_editor, 'some +  some ', delay=delay)
    qtbot.keyPress(code_editor, Qt.Key_Left)
    qtbot.keyPress(code_editor, Qt.Key_Left)
    qtbot.keyPress(code_editor, Qt.Key_Left)
    qtbot.keyPress(code_editor, Qt.Key_Left)
    qtbot.keyPress(code_editor, Qt.Key_Left)
    qtbot.keyPress(code_editor, Qt.Key_Backspace)
    qtbot.wait(2000)
    assert completion.isHidden()

    code_editor.toggle_code_snippets(True)


@pytest.mark.order(1)
@flaky(max_runs=5)
def test_automatic_completions_widget_visible(completions_codeeditor, qtbot):
    """
    Test on-the-fly completions when the widget is visible and the Backspace
    key is pressed.

    Regression test for PR #12710
    """
    code_editor, _ = completions_codeeditor
    completion = code_editor.completion_widget
    code_editor.toggle_code_snippets(False)

    code_editor.set_text('import math')
    cursor = code_editor.textCursor()
    code_editor.moveCursor(cursor.End)
    qtbot.keyPress(code_editor, Qt.Key_Enter, delay=300)  # newline

    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000):
        qtbot.keyClicks(code_editor, 'math.acosh', delay=300)

    assert completion.isVisible()

    qtbot.keyPress(code_editor, Qt.Key_Backspace, delay=300)
    qtbot.wait(500)
    assert completion.isVisible()

    qtbot.keyPress(code_editor, Qt.Key_Backspace, delay=300)
    qtbot.wait(500)
    assert completion.isVisible()

    code_editor.toggle_code_snippets(True)


if __name__ == '__main__':
    pytest.main(['test_completion.py', '--run-slow'])



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# ----------------------------------------------------------------------------
"""Tests for plugin config dialog."""

# Standard library imports
from unittest.mock import Mock

# Test library imports
from qtpy.QtCore import Signal
from qtpy.QtWidgets import QMainWindow
import pytest

# Local imports
from spyder.plugins.completion.plugin import CompletionPlugin
from spyder.plugins.preferences.tests.conftest import config_dialog  # noqa


class MainWindowMock(QMainWindow):
    sig_setup_finished = Signal()

    def __init__(self, parent):
        super(MainWindowMock, self).__init__(parent)
        self.statusbar = Mock()
        self.console = Mock()


@pytest.mark.parametrize(
    'config_dialog',
    [[MainWindowMock, [], [CompletionPlugin]]],
    indirect=True)
def test_config_dialog(config_dialog):
    expected_titles = {'General', 'Snippets', 'Linting', 'Introspection',
                       'Code style and formatting', 'Docstring style',
                       'Advanced', 'Other languages'}

    configpage = config_dialog.get_page()
    assert configpage
    tabs = configpage.tabs
    for i in range(0, tabs.count()):
        tab_text = tabs.tabText(i)
        assert tab_text in expected_titles
    configpage.save_to_conf()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Shell Interpreter"""

import sys
import atexit
import threading
import ctypes
import os
import re
import os.path as osp
import pydoc
from code import InteractiveConsole

from spyder_kernels.utils.dochelpers import isdefined

# Local imports:
from spyder.utils import encoding, programs
from spyder.py3compat import is_text_string
from spyder.utils.misc import remove_backslashes, getcwd_or_home

# Force Python to search modules in the current directory first:
sys.path.insert(0, '')


def guess_filename(filename):
    """Guess filename"""
    if osp.isfile(filename):
        return filename
    if not filename.endswith('.py'):
        filename += '.py'
    for path in [getcwd_or_home()] + sys.path:
        fname = osp.join(path, filename)
        if osp.isfile(fname):
            return fname
        elif osp.isfile(fname+'.py'):
            return fname+'.py'
        elif osp.isfile(fname+'.pyw'):
            return fname+'.pyw'
    return filename

class Interpreter(InteractiveConsole, threading.Thread):
    """Interpreter, executed in a separate thread"""
    p1 = ">>> "
    p2 = "... "
    def __init__(self, namespace=None, exitfunc=None,
                 Output=None, WidgetProxy=None, debug=False):
        """
        namespace: locals send to InteractiveConsole object
        commands: list of commands executed at startup
        """
        InteractiveConsole.__init__(self, namespace)
        threading.Thread.__init__(self)

        self._id = None

        self.exit_flag = False
        self.debug = debug

        # Execution Status
        self.more = False

        if exitfunc is not None:
            atexit.register(exitfunc)

        self.namespace = self.locals
        self.namespace['__name__'] = '__main__'
        self.namespace['execfile'] = self.execfile
        self.namespace['runfile'] = self.runfile
        self.namespace['raw_input'] = self.raw_input_replacement
        self.namespace['help'] = self.help_replacement

        # Capture all interactive input/output
        self.initial_stdout = sys.stdout
        self.initial_stderr = sys.stderr
        self.initial_stdin = sys.stdin

        # Create communication pipes
        pr, pw = os.pipe()
        self.stdin_read = os.fdopen(pr, "r")
        self.stdin_write = os.fdopen(pw, "wb", 0)
        self.stdout_write = Output()
        self.stderr_write = Output()

        self.input_condition = threading.Condition()
        self.widget_proxy = WidgetProxy(self.input_condition)

        self.redirect_stds()


    #------ Standard input/output
    def redirect_stds(self):
        """Redirects stds"""
        if not self.debug:
            sys.stdout = self.stdout_write
            sys.stderr = self.stderr_write
            sys.stdin = self.stdin_read

    def restore_stds(self):
        """Restore stds"""
        if not self.debug:
            sys.stdout = self.initial_stdout
            sys.stderr = self.initial_stderr
            sys.stdin = self.initial_stdin

    def raw_input_replacement(self, prompt=''):
        """For raw_input builtin function emulation"""
        self.widget_proxy.wait_input(prompt)
        self.input_condition.acquire()
        while not self.widget_proxy.data_available():
            self.input_condition.wait()
        inp = self.widget_proxy.input_data
        self.input_condition.release()
        return inp

    def help_replacement(self, text=None, interactive=False):
        """For help builtin function emulation"""
        if text is not None and not interactive:
            return pydoc.help(text)
        elif text is None:
            pyver = "%d.%d" % (sys.version_info[0], sys.version_info[1])
            self.write("""
Welcome to Python %s!  This is the online help utility.

If this is your first time using Python, you should definitely check out
the tutorial on the Internet at https://www.python.org/about/gettingstarted/

Enter the name of any module, keyword, or topic to get help on writing
Python programs and using Python modules.  To quit this help utility and
return to the interpreter, just type "quit".

To get a list of available modules, keywords, or topics, type "modules",
"keywords", or "topics".  Each module also comes with a one-line summary
of what it does; to list the modules whose summaries contain a given word
such as "spam", type "modules spam".
""" % pyver)
        else:
            text = text.strip()
            try:
                eval("pydoc.help(%s)" % text)
            except (NameError, SyntaxError):
                print("no Python documentation found for '%r'" % text) # spyder: test-skip
        self.write(os.linesep)
        self.widget_proxy.new_prompt("help> ")
        inp = self.raw_input_replacement()
        if inp.strip():
            self.help_replacement(inp, interactive=True)
        else:
            self.write("""
You are now leaving help and returning to the Python interpreter.
If you want to ask for help on a particular object directly from the
interpreter, you can type "help(object)".  Executing "help('string')"
has the same effect as typing a particular string at the help> prompt.
""")

    def run_command(self, cmd, new_prompt=True):
        """Run command in interpreter"""
        if cmd == 'exit()':
            self.exit_flag = True
            self.write('\n')
            return
        # -- Special commands type I
        #    (transformed into commands executed in the interpreter)
        # ? command
        special_pattern = r"^%s (?:r\')?(?:u\')?\"?\'?([a-zA-Z0-9_\.]+)"
        run_match = re.match(special_pattern % 'run', cmd)
        help_match = re.match(r'^([a-zA-Z0-9_\.]+)\?$', cmd)
        cd_match = re.match(r"^\!cd \"?\'?([a-zA-Z0-9_ \.]+)", cmd)
        if help_match:
            cmd = 'help(%s)' % help_match.group(1)
        # run command
        elif run_match:
            filename = guess_filename(run_match.groups()[0])
            cmd = "%runfile " + repr(remove_backslashes(filename))
        # !cd system command
        elif cd_match:
            cmd = 'import os; os.chdir(r"%s")' % cd_match.groups()[0].strip()
        # -- End of Special commands type I

        # -- Special commands type II
        #    (don't need code execution in interpreter)
        xedit_match = re.match(special_pattern % 'xedit', cmd)
        edit_match = re.match(special_pattern % 'edit', cmd)
        clear_match = re.match(r"^clear ([a-zA-Z0-9_, ]+)", cmd)
        # (external) edit command
        if xedit_match:
            filename = guess_filename(xedit_match.groups()[0])
            self.widget_proxy.edit(filename, external_editor=True)
        # local edit command
        elif edit_match:
            filename = guess_filename(edit_match.groups()[0])
            if osp.isfile(filename):
                self.widget_proxy.edit(filename)
            else:
                self.stderr_write.write(
                                "No such file or directory: %s\n" % filename)
        # remove reference (equivalent to MATLAB's clear command)
        elif clear_match:
            varnames = clear_match.groups()[0].replace(' ', '').split(',')
            for varname in varnames:
                try:
                    self.namespace.pop(varname)
                except KeyError:
                    pass
        # Execute command
        elif cmd.startswith('!'):
            # System ! command
            pipe = programs.run_shell_command(cmd[1:])
            txt_out = encoding.transcode( pipe.stdout.read().decode() )
            txt_err = encoding.transcode( pipe.stderr.read().decode().rstrip() )
            if txt_err:
                self.stderr_write.write(txt_err)
            if txt_out:
                self.stdout_write.write(txt_out)
            self.stdout_write.write('\n')
            self.more = False
        # -- End of Special commands type II
        else:
            # Command executed in the interpreter
#            self.widget_proxy.set_readonly(True)
            self.more = self.push(cmd)
#            self.widget_proxy.set_readonly(False)

        if new_prompt:
            self.widget_proxy.new_prompt(self.p2 if self.more else self.p1)
        if not self.more:
            self.resetbuffer()

    def run(self):
        """Wait for input and run it"""
        while not self.exit_flag:
            self.run_line()

    def run_line(self):
        line = self.stdin_read.readline()
        if self.exit_flag:
            return
        # Remove last character which is always '\n':
        self.run_command(line[:-1])

    def get_thread_id(self):
        """Return thread id"""
        if self._id is None:
            for thread_id, obj in list(threading._active.items()):
                if obj is self:
                    self._id = thread_id
        return self._id

    def raise_keyboard_interrupt(self):
        if self.isAlive():
            ctypes.pythonapi.PyThreadState_SetAsyncExc(self.get_thread_id(),
                                           ctypes.py_object(KeyboardInterrupt))
            return True
        else:
            return False


    def closing(self):
        """Actions to be done before restarting this interpreter"""
        pass

    def execfile(self, filename):
        """Exec filename"""
        source = open(filename, 'r').read()
        try:
            try:
                name = filename.encode('ascii')
            except UnicodeEncodeError:
                name = '<executed_script>'
            code = compile(source, name, "exec")
        except (OverflowError, SyntaxError):
            InteractiveConsole.showsyntaxerror(self, filename)
        else:
            self.runcode(code)

    def runfile(self, filename, args=None):
        """
        Run filename
        args: command line arguments (string)
        """
        if args is not None and not is_text_string(args):
            raise TypeError("expected a character buffer object")
        self.namespace['__file__'] = filename
        sys.argv = [filename]
        if args is not None:
            for arg in args.split():
                sys.argv.append(arg)
        self.execfile(filename)
        sys.argv = ['']
        self.namespace.pop('__file__')

    def eval(self, text):
        """
        Evaluate text and return (obj, valid)
        where *obj* is the object represented by *text*
        and *valid* is True if object evaluation did not raise any exception
        """
        assert is_text_string(text)
        try:
            return eval(text, self.locals), True
        except:
            return None, False

    def is_defined(self, objtxt, force_import=False):
        """Return True if object is defined"""
        return isdefined(objtxt, force_import=force_import,
                         namespace=self.locals)

    #===========================================================================
    # InteractiveConsole API
    #===========================================================================
    def push(self, line):
        """
        Push a line of source text to the interpreter

        The line should not have a trailing newline; it may have internal
        newlines. The line is appended to a buffer and the interpreter’s
        runsource() method is called with the concatenated contents of the
        buffer as source. If this indicates that the command was executed
        or invalid, the buffer is reset; otherwise, the command is incomplete,
        and the buffer is left as it was after the line was appended.
        The return value is True if more input is required, False if the line
        was dealt with in some way (this is the same as runsource()).
        """
        return InteractiveConsole.push(self, "#coding=utf-8\n" + line)

    def resetbuffer(self):
        """Remove any unhandled source text from the input buffer"""
        InteractiveConsole.resetbuffer(self)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Switcher Item Widget."""

# Standard library imports
import os
import sys

# Third party imports
from qtpy.QtCore import (QSize, Qt)
from qtpy.QtGui import QStandardItem, QTextDocument

# Local imports
from spyder.config.utils import is_ubuntu
from spyder.py3compat import to_text_string


class SwitcherBaseItem(QStandardItem):
    """Base List Item."""

    _PADDING = 5
    _WIDTH = 400
    _TEMPLATE = None

    def __init__(self, parent=None, styles=None, use_score=True):
        """Create basic List Item."""
        super().__init__()

        # Style
        self._width = self._WIDTH
        self._padding = self._PADDING
        self._styles = styles if styles else {}
        self._action_item = False
        self._score = -1
        self._height = self._get_height()
        self._use_score = use_score

        # Setup
        # self._height is a float from QSizeF but
        # QSize() expects a QSize or (int, int) as parameters
        self.setSizeHint(QSize(0, int(self._height)))

    def _render_text(self):
        """Render the html template for this item."""
        raise NotImplementedError

    def _set_rendered_text(self):
        """Set the rendered html template as text of this item."""
        self.setText(self._render_text())

    def _set_styles(self):
        """Set the styles for this item."""
        raise NotImplementedError

    def _get_height(self):
        """
        Return the expected height of this item's text, including
        the text margins.
        """
        raise NotImplementedError

    # ---- API
    def set_width(self, value):
        """Set the content width."""
        self._width = value - (self._padding * 3)
        self._set_rendered_text()

    def get_width(self):
        """Return the content width."""
        return self._width

    def get_height(self):
        """Return the content height."""
        return self._height

    def get_score(self):
        """Return the fuzzy matchig score."""
        return self._score

    def set_score(self, value):
        """Set the search text fuzzy match score."""
        if self._use_score:
            self._score = value
            self._set_rendered_text()

    def is_action_item(self):
        """Return whether the item is of action type."""
        return bool(self._action_item)

    # ---- Qt overrides
    def refresh(self):
        """Override Qt."""
        super().refresh()
        self._set_styles()
        self._set_rendered_text()


class SwitcherSeparatorItem(SwitcherBaseItem):
    """
    Separator Item represented as <hr>.

    Based on HTML delegate.

    See: https://doc.qt.io/qt-5/richtext-html-subset.html
    """

    _SEPARATOR = '_'
    _STYLE_ATTRIBUTES = ['color', 'font_size']
    _TEMPLATE = \
        u'''<table cellpadding="{padding}" cellspacing="0" width="{width}"
                  height="{height}" border="0">
  <tr><td valign="top" align="center"><hr></td></tr>
</table>'''

    def __init__(self, parent=None, styles=None):
        """Separator Item represented as <hr>."""
        super().__init__(parent=parent, styles=styles)
        self.setFlags(Qt.NoItemFlags)
        self._set_rendered_text()

    # ---- Helpers
    def _set_styles(self):
        """Set the styles for this item."""
        for attr in self._STYLE_ATTRIBUTES:
            if attr not in self._styles:
                self._styles[attr] = self._STYLES[attr]

        rich_font = self._styles['font_size']

        if sys.platform == 'darwin':
            font_size = rich_font
        elif os.name == 'nt':
            font_size = rich_font
        elif is_ubuntu():
            font_size = rich_font - 2
        else:
            font_size = rich_font - 2

        self._styles['font_size'] = font_size

    def _render_text(self):
        """Render the html template for this item."""
        padding = self._padding
        width = self._width
        height = self.get_height()
        text = self._TEMPLATE.format(width=width, height=height,
                                     padding=padding, **self._styles)
        return text

    def _get_height(self):
        """
        Return the expected height of this item's text, including
        the text margins.
        """
        doc = QTextDocument()
        doc.setHtml('<hr>')
        doc.setDocumentMargin(self._PADDING)
        return doc.size().height()


class SwitcherItem(SwitcherBaseItem):
    """
    Switcher item with title, description, shortcut and section.

    SwitcherItem: [title description    <shortcut> section]

    Based on HTML delegate.
    See: https://doc.qt.io/qt-5/richtext-html-subset.html
    """

    _FONT_SIZE = 10
    _STYLE_ATTRIBUTES = ['title_color', 'description_color', 'section_color',
                         'shortcut_color', 'title_font_size',
                         'description_font_size', 'section_font_size',
                         'shortcut_font_size']
    _TEMPLATE = u'''
<table width="{width}" max_width="{width}" height="{height}"
                          cellpadding="{padding}">
  <tr>
    <td valign="middle">
      <span style="color:{title_color};font-size:{title_font_size}pt">
        {title}
      </span>
      &nbsp;&nbsp;
      <em>
        <span
         style="color:{description_color};font-size:{description_font_size}pt">
          <span>{description}</span>
        </span>
      </em>
    </td>
    <td valign="middle" align="right" float="right">
      <span style="color:{shortcut_color};font-size:{shortcut_font_size}pt">
         <code><i>{shortcut}</i></code>
      </span>&nbsp;
      <span style="color:{section_color};font-size:{section_font_size}pt">
         {section}
      </span>
    </td>
  </tr>
</table>'''

    def __init__(self, parent=None, icon=None, title=None, description=None,
                 shortcut=None, section=None, data=None, tool_tip=None,
                 action_item=False, styles=None, score=-1, use_score=True):
        """Switcher item with title, description, shortcut and section."""
        super().__init__(parent=parent, styles=styles, use_score=use_score)

        self._title = title if title else ''
        self._rich_title = ''
        self._shortcut = shortcut if shortcut else ''
        self._description = description if description else ''
        self._section = section if section else ''
        self._icon = icon
        self._data = data
        self._score = score
        self._action_item = action_item

        # Section visibility is computed by the setup_sections method of the
        # switcher.
        self._section_visible = False

        # Setup
        self.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable)

        if icon:
            self.setIcon(icon)
            # TODO: Change fixed icon size value
            self._icon_width = 20
        else:
            self._icon_width = 0

        self._set_styles()
        self._set_rendered_text()

    # ---- Helpers
    def _render_text(self, title=None, description=None, section=None):
        """Render the html template for this item."""
        if self._rich_title:
            title = self._rich_title
        else:
            title = title if title else self._title

        # TODO: Based on width this should elide/shorten
        description = description if description else self._description

        if self._section_visible:
            section = section if section else self._section
        else:
            section = ''

        padding = self._PADDING
        width = int(self._width - self._icon_width)
        height = int(self.get_height())
        self.setSizeHint(QSize(width, height))

        shortcut = '&lt;' + self._shortcut + '&gt;' if self._shortcut else ''

        title = to_text_string(title, encoding='utf-8')
        section = to_text_string(section, encoding='utf-8')
        description = to_text_string(description, encoding='utf-8')
        shortcut = to_text_string(shortcut, encoding='utf-8')

        text = self._TEMPLATE.format(width=width, height=height, title=title,
                                     section=section, description=description,
                                     padding=padding, shortcut=shortcut,
                                     **self._styles)
        return text

    def _set_styles(self):
        """Set the styles for this item."""
        for attr in self._STYLE_ATTRIBUTES:
            if attr not in self._styles:
                self._styles[attr] = self._STYLES[attr]

    def _get_height(self):
        """
        Return the expected height of this item's text, including
        the text margins.
        """
        doc = QTextDocument()
        try:
            doc.setHtml('<span style="font-size:{}pt">Title</span>'
                        .format(self._styles['title_font_size']))
        except KeyError:
            doc.setHtml('<span>Title</span>')
        doc.setDocumentMargin(self._PADDING)
        return doc.size().height()

    # ---- API
    def set_icon(self, icon):
        """Set the QIcon for the list item."""
        self._icon = icon
        self.setIcon(icon)

    def get_icon(self):
        """Return the QIcon for the list item."""
        return self._icon

    def set_title(self, value):
        """Set the main text (title) of the item."""
        self._title = value
        self._set_rendered_text()

    def get_title(self):
        """Return the the main text (title) of the item."""
        return self._title

    def set_rich_title(self, value):
        """Set the rich title version (filter highlight) of the item."""
        self._rich_title = value
        self._set_rendered_text()

    def get_rich_title(self):
        """Return the rich title version (filter highlight) of the item."""
        return self._rich_title

    def set_description(self, value):
        """Set the item description text."""
        self._description = value
        self._set_rendered_text()

    def get_description(self):
        """Return the item description text."""
        return self._description

    def set_shortcut(self, value):
        """Set the shortcut for the item action."""
        self._shortcut = value
        self._set_rendered_text()

    def get_shortcut(self, value):
        """Return the shortcut for the item action."""
        return self._shortcut

    def set_tooltip(self, value):
        """Set the tooltip text for the item."""
        super().setTooltip(value)

    def set_data(self, value):
        """Set the additional data associated to the item."""
        self._data = value

    def get_data(self):
        """Return the additional data associated to the item."""
        return self._data

    def set_section(self, value):
        """Set the item section name."""
        self._section = value
        self._set_rendered_text()

    def get_section(self):
        """Return the item section name."""
        return self._section

    def set_section_visible(self, value):
        """Set visibility of the item section."""
        self._section_visible = value
        self._set_rendered_text()

    def set_action_item(self, value):
        """Enable/disable the action type for the item."""
        self._action_item = value
        self._set_rendered_text()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

# Third party imports
from qtpy.QtWidgets import QPlainTextEdit

import pytest

# Local imports
from spyder.widgets import mixins


class BaseWidget(QPlainTextEdit, mixins.BaseEditMixin):
    pass


# --- Fixtures
# -----------------------------------------------------------------------------


@pytest.fixture
def mixinsbot(qtbot):
    widget = BaseWidget()
    qtbot.addWidget(widget)
    widget.show()
    return qtbot, widget


# --- Tests
# -----------------------------------------------------------------------------
@pytest.mark.parametrize(
    "test_object", [(u"test.attr = np.linalg.norm()\n",
                     [(None, 'test'),
                      (4,    'test'),
                      (1,    'test.attr'),
                      (4,    'test.attr'),
                      (1,     None),
                      (1,     'np'),
                      (4,     'np.linalg'),
                      (6,     'np.linalg'),
                      (1,     'np.linalg.norm'),
                      (4,     'np.linalg.norm'),
                      (1,     None)
                      ]),
                    (u"test$attr = np$linalg$norm()\n",
                     [(None, 'test'),
                      (4,    'test'),
                      (1,    'test$attr'),
                      (4,    'test$attr'),
                      (1,     None),
                      (1,     'np'),
                      (4,     'np$linalg'),
                      (6,     'np$linalg'),
                      (1,     'np$linalg$norm'),
                      (4,     'np$linalg$norm'),
                      (1,     None)
                      ])])
def test_get_current_object(mixinsbot, test_object):
    """
    Test that we can get the current object to get help for the console.

    For spyder-ide/spyder#11821
    """
    qtbot, widget = mixinsbot
    get_current_object = widget.get_current_object

    code, matches = test_object
    widget.setPlainText(code)
    cursor = widget.textCursor()
    cursor.setPosition(widget.get_position('sof'))

    for move, match in matches:
        if move is not None:
            widget.move_cursor(move)
        current_word = get_current_object()
        assert current_word == match


def test_get_current_word(mixinsbot):
    """
    Test that we can get the current word to get help for the editor.

    For spyder-ide/spyder#11267
    """
    qtbot, widget = mixinsbot
    get_current_word = widget.get_current_word

    code = (u'def foo(a, b, c)\n')
    widget.setPlainText(code)
    cursor = widget.textCursor()
    cursor.setPosition(widget.get_position('sof'))
    widget.move_cursor(7)
    current_word = get_current_word()
    assert current_word is None
    current_word = get_current_word(help_req=True)
    assert current_word == 'foo'


def test_get_unicode_regexp(mixinsbot):
    """
    Test that we can search with regexp's containing unicode
    characters.

    For spyder-ide/spyder#6812.
    """
    qtbot, widget = mixinsbot
    get = widget.get_number_matches

    # CodeEditor and findreplace texts are handled in
    # unicode by PyQt5 in Python 2
    code = (u'print("И")\n'
            u'foo("И")')
    widget.setPlainText(code)
    cursor = widget.textCursor()
    cursor.setPosition(widget.get_position('sof'))
    assert widget.find_text(u't.*И', regexp=True)
    assert get(u't.*И', source_text=code, regexp=True) == 1


def test_get_number_matches(mixinsbot):
    # Test get_number_matches().
    qtbot, widget = mixinsbot
    get = widget.get_number_matches

    code = ('class C():\n'
            '    def __init__(self):\n'
            '        pass\n'
            '    def f(self, a, b):\n'
            '        pass\n')

    # Empty pattern.
    assert get('') == 0

    # No case, no regexp.
    assert get('self', source_text=code) == 2
    assert get('c', source_text=code) == 2

    # Case, no regexp.
    assert get('self', source_text=code, case=True) == 2
    assert get('c', source_text=code, case=True) == 1

    # No case, regexp.
    assert get('e[a-z]?f', source_text=code, regexp=True) == 4
    assert get('e[A-Z]?f', source_text=code, regexp=True) == 4

    # Case, regexp.
    assert get('e[a-z]?f', source_text=code, case=True, regexp=True) == 4
    assert get('e[A-Z]?f', source_text=code, case=True, regexp=True) == 2

    # No case, regexp, word
    assert get('e[a-z]?f', source_text=code, regexp=True, word=True) == 0
    assert get('e[A-Z]?f', source_text=code, regexp=True, word=True) == 0

    # Case, regexp, word
    assert get('e[a-z]?f', source_text=code, case=True, regexp=True,
               word=True) == 0
    assert get('e[A-Z]?f', source_text=code, case=True, regexp=True,
               word=True) == 0

    # spyder-ide/spyder#5680.
    assert get('(', source_text=code) == 3
    assert get('(', source_text=code, case=True) == 3
    assert get('(', source_text=code, regexp=True) is None
    assert get('(', source_text=code, case=True, regexp=True) is None

    # spyder-ide/spyder#7960.
    assert get('a', source_text=code) == 4
    assert get('a', source_text=code, case=True) == 4
    assert get('a', source_text=code, regexp=True) == 4
    assert get('a', source_text=code, case=True, regexp=True) == 4
    assert get('a', source_text=code, case=True, regexp=True, word=True) == 1
    assert get('a', source_text=code, regexp=True, word=True) == 1
    assert get('a', source_text=code, case=True, word=True) == 1


def test_get_match_number(mixinsbot):
    # Test get_match_number().
    qtbot, widget = mixinsbot
    get = widget.get_match_number

    code = ('class C():\n'
            '    def __init__(self):\n'
            '        pass\n'
            '    def f(self, a, b):\n'
            '        pass\n')
    widget.setPlainText(code)
    cursor = widget.textCursor()
    cursor.setPosition(widget.get_position('sof'))

    # Empty pattern.
    assert get('') == 0

    # spyder-ide/spyder#5680.
    widget.find_text('(')
    assert get('(') == 1

    # First occurrence.
    widget.find_text('self')
    assert get('self') == 1
    assert get('self', case=True) == 1

    # Second occurrence.
    widget.find_text('pass')
    widget.find_text('self')
    assert get('self') == 2
    assert get('self', case=True) == 2


def test_get_number_with_words(mixinsbot):
    """
    Test that find count honours the word setting.

    Dedicated test for spyder-ide/spyder#7960.
    """
    qtbot, widget = mixinsbot
    getn = widget.get_number_matches
    getm = widget.get_match_number

    code = ('word\n'
            'words\n'
            'Word\n'
            'Words\n'
            'sword\n'
            'word\n')

    widget.setPlainText(code)
    cursor = widget.textCursor()
    cursor.setPosition(widget.get_position('sof'))

    widget.find_text('Word')
    # Six instances of the string in total
    assert getn('Word', source_text=code) == 6
    # Found the first
    assert getm('Word') == 1
    # Two with same casing
    assert getn('Word', source_text=code, case=True) == 2
    # Three [Ww]ord in total
    assert getn('Word', source_text=code, word=True) == 3
    # But only one Word
    assert getn('Word', source_text=code, word=True, case=True) == 1
    # Find next Word case sensitive
    widget.find_text('Word', case=True)
    # This also moves to third row and first instance of Word (case sensitive)
    assert getm('Word', case=True) == 1
    # So this should be the third instance if case not considered
    assert getm('Word') == 3
    # But the second (out of the three) which is [Ww]ord
    assert getm('Word', word=True) == 2
    # Find something on next line just to progress
    widget.find_text('Words')
    # Find next [Ww]ord word-sensitive
    widget.find_text('Word', word=True)
    # This should be the sixth instance if case not considered
    assert getm('Word') == 6
    # But the third (out of the three) which is [Ww]ord
    assert getm('Word', word=True) == 3



# -*- coding: utf-8 -*-
#
# Copyright (c) 2009- Spyder Project Contributors
#
# Distributed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
API utilities.
"""


def get_class_values(cls):
    """
    Get the attribute values for the class enumerations used in our API.

    Idea from: https://stackoverflow.com/a/17249228/438386
    """
    return [v for (k, v) in cls.__dict__.items() if k[:1] != '_']


class PrefixNode:
    """Utility class used to represent a prefixed string tuple."""

    def __init__(self, path=None):
        self.children = {}
        self.path = path

    def __iter__(self):
        prefix = [((self.path,), self)]
        while prefix != []:
            current_prefix, node = prefix.pop(0)
            prefix += [(current_prefix + (c,), node.children[c])
                       for c in node.children]
            yield current_prefix

    def add_path(self, path):
        prefix, *rest = path
        if prefix not in self.children:
            self.children[prefix] = PrefixNode(prefix)

        if len(rest) > 0:
            child = self.children[prefix]
            child.add_path(rest)


class PrefixedTuple(PrefixNode):
    """Utility class to store and iterate over prefixed string tuples."""

    def __iter__(self):
        for key in self.children:
            child = self.children[key]
            for prefix in child:
                yield prefix


class classproperty(property):
    """
    Decorator to declare class constants as properties that require additional
    computation.

    Taken from: https://stackoverflow.com/a/7864317/438386
    """

    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""Tests for code formatting using CodeEditor instances."""

# Third party imports
import pytest
from qtpy.QtGui import QTextCursor

# Local imports
from spyder.config.manager import CONF
from spyder.plugins.editor.widgets.codeeditor.tests.conftest import (
    autopep8,
    black,
    yapf,
    get_formatter_values
)


# ---- Tests
@pytest.mark.order(1)
@pytest.mark.parametrize('formatter', [autopep8, yapf, black])
@pytest.mark.parametrize('newline', ['\r\n', '\r', '\n'])
def test_document_formatting(formatter, newline, completions_codeeditor,
                             qtbot):
    """Validate text autoformatting via autopep8, yapf or black."""
    code_editor, completion_plugin = completions_codeeditor
    text, expected = get_formatter_values(formatter, newline)

    # Set formatter
    CONF.set(
        'completions',
        ('provider_configuration', 'lsp', 'values', 'formatting'),
        formatter
    )
    completion_plugin.after_configuration_update([])
    qtbot.wait(2000)

    # Set text in editor
    code_editor.set_text(text)

    # Assert eols are the expected ones
    assert code_editor.get_line_separator() == newline

    # Notify changes
    with qtbot.waitSignal(
            code_editor.completions_response_signal, timeout=30000):
        code_editor.document_did_change()

    # Perform formatting
    with qtbot.waitSignal(
            code_editor.completions_response_signal, timeout=30000):
        code_editor.format_document()

    # Wait for text to be formatted
    qtbot.wait(2000)

    assert code_editor.get_text_with_eol() == expected


@pytest.mark.order(1)
@pytest.mark.parametrize('formatter', [autopep8, yapf, black])
@pytest.mark.parametrize('newline', ['\r\n', '\r', '\n'])
def test_document_range_formatting(formatter, newline, completions_codeeditor,
                                   qtbot):
    """Validate text range autoformatting."""
    code_editor, completion_plugin = completions_codeeditor
    text, expected = get_formatter_values(formatter, newline, range_fmt=True)

    # This is broken in PyLSP 1.5.0. We need to investigate why.
    if formatter == 'yapf':
        return

    # Set formatter
    CONF.set(
        'completions',
        ('provider_configuration', 'lsp', 'values', 'formatting'),
        formatter
    )
    completion_plugin.after_configuration_update([])
    qtbot.wait(2000)

    # Set text in editor
    code_editor.set_text(text)

    # Assert eols are the expected ones
    assert code_editor.get_line_separator() == newline

    # Notify changes
    with qtbot.waitSignal(
            code_editor.completions_response_signal, timeout=30000):
        code_editor.document_did_change()

    # Select region to format
    code_editor.go_to_line(12)
    cursor = code_editor.textCursor()
    start = code_editor.get_position_line_number(11, -1)
    end = code_editor.get_position_line_number(27, 0)
    cursor.setPosition(start)
    cursor.setPosition(end, QTextCursor.KeepAnchor)
    cursor.movePosition(QTextCursor.EndOfBlock,
                        QTextCursor.KeepAnchor)
    code_editor.setTextCursor(cursor)

    # Perform formatting
    with qtbot.waitSignal(
            code_editor.completions_response_signal, timeout=30000):
        code_editor.format_document_range()

    # Wait to text to be formatted
    qtbot.wait(2000)
    assert code_editor.get_text_with_eol() == expected


@pytest.mark.order(1)
@pytest.mark.parametrize('formatter', [autopep8, black])
def test_max_line_length(formatter, completions_codeeditor, qtbot):
    """Validate autoformatting with a different value of max_line_length."""
    code_editor, completion_plugin = completions_codeeditor
    text, expected = get_formatter_values(
        formatter, newline='\n', max_line=True)
    max_line_length = 20

    # Set formatter and max line length options
    CONF.set(
        'completions',
        ('provider_configuration', 'lsp', 'values', 'formatting'),
        formatter
    )
    CONF.set(
        'completions',
        ('provider_configuration', 'lsp', 'values',
         'pycodestyle/max_line_length'),
        max_line_length
    )
    completion_plugin.after_configuration_update([])
    qtbot.wait(2000)

    # Set text in editor
    code_editor.set_text(text)

    # Notify changes
    with qtbot.waitSignal(
            code_editor.completions_response_signal, timeout=30000):
        code_editor.document_did_change()

    # Perform formatting
    with qtbot.waitSignal(
            code_editor.completions_response_signal, timeout=30000):
        code_editor.format_document()

    # Wait for text to be formatted
    qtbot.wait(2000)

    assert code_editor.get_text_with_eol() == expected



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
spyder.api.plugins
==================

Here, 'plugins' are Qt objects that can make changes to Spyder's main window
and call/connect to other plugins directly.

There are two types of plugins available:

1. SpyderPluginV2 is a plugin that does not create a new dock/pane on Spyder's
   main window. Note: SpyderPluginV2 will be renamed to SpyderPlugin once the
   migration to the new API is finished

2. SpyderDockablePlugin is a plugin that does create a new dock/pane on
   Spyder's main window.
"""

from .enum import Plugins, DockablePlugins  # noqa
from .new_api import SpyderDockablePlugin, SpyderPluginV2  # noqa



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
General entry in Preferences.

For historical reasons (dating back to Spyder 2) the main class here is called
`MainConfigPage` and its associated entry in our config system is called
`main`.
"""

import traceback
import sys

from qtpy.compat import from_qvariant
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (QApplication, QButtonGroup, QGridLayout, QGroupBox,
                            QHBoxLayout, QLabel, QMessageBox, QVBoxLayout,
                            QWidget)

from spyder.config.base import (_, DISABLED_LANGUAGES, LANGUAGE_CODES,
                                is_conda_based_app, save_lang_conf)
from spyder.api.plugins import Plugins
from spyder.api.preferences import PluginConfigPage
from spyder.py3compat import to_text_string

HDPI_QT_PAGE = "https://doc.qt.io/qt-5/highdpi.html"


class ApplicationConfigPage(PluginConfigPage):

    def __init__(self, plugin, parent):
        super().__init__(plugin, parent)
        self.pre_apply_callback = self.perform_checks

    def setup_page(self):
        newcb = self.create_checkbox

        # --- Advanced
        # Remove disabled languages
        language_codes = LANGUAGE_CODES.copy()
        for lang in DISABLED_LANGUAGES:
            language_codes.pop(lang)

        languages = language_codes.items()
        language_choices = sorted([(val, key) for key, val in languages])
        language_combo = self.create_combobox(_('Language:'),
                                              language_choices,
                                              'interface_language',
                                              restart=True)

        opengl_options = ['Automatic', 'Desktop', 'Software', 'GLES']
        opengl_choices = list(zip(opengl_options,
                                  [c.lower() for c in opengl_options]))
        opengl_combo = self.create_combobox(_('Rendering engine:'),
                                            opengl_choices,
                                            'opengl',
                                            restart=True)

        single_instance_box = newcb(_("Use a single instance"),
                                    'single_instance',
                                    tip=_("Set this to open external<br> "
                                          "Python files in an already running "
                                          "instance (Requires a restart)"))

        prompt_box = newcb(_("Prompt when exiting"), 'prompt_on_exit')
        popup_console_box = newcb(_("Show internal Spyder errors to report "
                                    "them to Github"), 'show_internal_errors')
        check_update_cb = newcb(
            _("Check for updates on startup"),
            'check_updates_on_startup',
            section='update_manager'
        )
        stable_only_cb = newcb(
            _("Check for stable releases only"),
            'check_stable_only',
            section='update_manager'
        )

        # Decide if it's possible to activate or not single instance mode
        # ??? Should we allow multiple instances for macOS?
        if sys.platform == 'darwin' and is_conda_based_app():
            self.set_option("single_instance", True)
            single_instance_box.setEnabled(False)

        comboboxes_advanced_layout = QHBoxLayout()
        cbs_adv_grid = QGridLayout()
        cbs_adv_grid.addWidget(language_combo.label, 0, 0)
        cbs_adv_grid.addWidget(language_combo.combobox, 0, 1)
        cbs_adv_grid.addWidget(opengl_combo.label, 1, 0)
        cbs_adv_grid.addWidget(opengl_combo.combobox, 1, 1)
        comboboxes_advanced_layout.addLayout(cbs_adv_grid)
        comboboxes_advanced_layout.addStretch(1)

        advanced_layout = QVBoxLayout()
        advanced_layout.addLayout(comboboxes_advanced_layout)
        advanced_layout.addWidget(single_instance_box)
        advanced_layout.addWidget(prompt_box)
        advanced_layout.addWidget(popup_console_box)
        advanced_layout.addWidget(check_update_cb)
        advanced_layout.addWidget(stable_only_cb)

        advanced_widget = QWidget()
        advanced_widget.setLayout(advanced_layout)

        # --- Panes
        interface_group = QGroupBox(_("Panes"))

        verttabs_box = newcb(_("Vertical tabs in panes"),
                             'vertical_tabs', restart=True)
        margin_box = newcb(_("Custom margin for panes:"),
                           'use_custom_margin')
        margin_spin = self.create_spinbox("", _("pixels"), 'custom_margin',
                                          default=0, min_=0, max_=30)
        margin_box.checkbox.toggled.connect(margin_spin.spinbox.setEnabled)
        margin_box.checkbox.toggled.connect(margin_spin.slabel.setEnabled)
        margin_spin.spinbox.setEnabled(self.get_option('use_custom_margin'))
        margin_spin.slabel.setEnabled(self.get_option('use_custom_margin'))

        cursor_box = newcb(_("Cursor blinking:"),
                           'use_custom_cursor_blinking')
        cursor_spin = self.create_spinbox(
            "", _("ms"),
            'custom_cursor_blinking',
            default=QApplication.cursorFlashTime(),
            min_=0, max_=5000, step=100)
        cursor_box.checkbox.toggled.connect(cursor_spin.spinbox.setEnabled)
        cursor_box.checkbox.toggled.connect(cursor_spin.slabel.setEnabled)
        cursor_spin.spinbox.setEnabled(
            self.get_option('use_custom_cursor_blinking'))
        cursor_spin.slabel.setEnabled(
            self.get_option('use_custom_cursor_blinking'))

        margins_cursor_layout = QGridLayout()
        margins_cursor_layout.addWidget(margin_box, 0, 0)
        margins_cursor_layout.addWidget(margin_spin.spinbox, 0, 1)
        margins_cursor_layout.addWidget(margin_spin.slabel, 0, 2)
        margins_cursor_layout.addWidget(cursor_box, 1, 0)
        margins_cursor_layout.addWidget(cursor_spin.spinbox, 1, 1)
        margins_cursor_layout.addWidget(cursor_spin.slabel, 1, 2)
        margins_cursor_layout.setColumnStretch(2, 100)

        # Layout interface
        interface_layout = QVBoxLayout()
        interface_layout.addWidget(verttabs_box)
        interface_layout.addLayout(margins_cursor_layout)
        interface_group.setLayout(interface_layout)

        if sys.platform == "darwin" and not is_conda_based_app():
            # To open files from Finder directly in Spyder.
            from spyder.utils.qthelpers import (register_app_launchservices,
                                                restore_launchservices)
            import applaunchservices as als

            def set_open_file(state):
                if state:
                    register_app_launchservices()
                else:
                    restore_launchservices()

            macOS_group = QGroupBox(_("macOS integration"))
            mac_open_file_box = newcb(
                _("Open files from Finder with Spyder"),
                'mac_open_file',
                tip=_("Register Spyder with the Launch Services"))
            mac_open_file_box.checkbox.toggled.connect(set_open_file)
            macOS_layout = QVBoxLayout()
            macOS_layout.addWidget(mac_open_file_box)
            if als.get_bundle_identifier() is None:
                # Disable setting
                mac_open_file_box.setDisabled(True)
                macOS_layout.addWidget(QLabel(
                    _('Launch Spyder with <code>python.app</code> to enable'
                      ' Apple event integrations.')))

            macOS_group.setLayout(macOS_layout)

        # --- Screen resolution Group (hidpi)
        screen_resolution_group = QGroupBox(_("Screen resolution"))
        screen_resolution_bg = QButtonGroup(screen_resolution_group)
        screen_resolution_label = QLabel(_("Configuration for high DPI "
                                           "screens<br><br>"
                                           "Please see "
                                           "<a href=\"{0}\">{0}</a><> "
                                           "for more information about "
                                           "these options (in "
                                           "English).").format(HDPI_QT_PAGE))
        screen_resolution_label.setWordWrap(True)
        screen_resolution_label.setOpenExternalLinks(True)

        self.normal_radio = self.create_radiobutton(
            _("Normal"),
            'normal_screen_resolution',
            button_group=screen_resolution_bg
        )
        auto_scale_radio = self.create_radiobutton(
            _("Enable auto high DPI scaling"),
            'high_dpi_scaling',
            button_group=screen_resolution_bg,
            tip=_("Set this for high DPI displays"),
            restart=True
        )

        self.custom_scaling_radio = self.create_radiobutton(
            _("Set a custom high DPI scaling"),
            'high_dpi_custom_scale_factor',
            button_group=screen_resolution_bg,
            tip=_("Set this for high DPI displays when auto scaling does not "
                  "work"),
            restart=True
        )

        self.custom_scaling_edit = self.create_lineedit(
            "",
            'high_dpi_custom_scale_factors',
            tip=_("Enter values for different screens separated by semicolons "
                  "';'.\n Float values are supported"),
            alignment=Qt.Horizontal,
            regex=r"[1-9]+(?:\.[0-9]*)(;[1-9]+(?:\.[0-9]*))*",
            restart=True
        )

        self.normal_radio.radiobutton.toggled.connect(
            self.custom_scaling_edit.textbox.setDisabled)
        auto_scale_radio.radiobutton.toggled.connect(
            self.custom_scaling_edit.textbox.setDisabled)
        self.custom_scaling_radio.radiobutton.toggled.connect(
            self.custom_scaling_edit.textbox.setEnabled)

        # Layout Screen resolution
        screen_resolution_layout = QVBoxLayout()
        screen_resolution_layout.addWidget(screen_resolution_label)

        screen_resolution_inner_layout = QGridLayout()
        screen_resolution_inner_layout.addWidget(self.normal_radio, 0, 0)
        screen_resolution_inner_layout.addWidget(
            auto_scale_radio.radiobutton, 1, 0)
        screen_resolution_inner_layout.addWidget(
            auto_scale_radio.radiobutton.help_label, 1, 1)
        screen_resolution_inner_layout.addWidget(
            self.custom_scaling_radio.radiobutton, 2, 0)
        screen_resolution_inner_layout.addWidget(
            self.custom_scaling_radio.radiobutton.help_label, 2, 1)
        screen_resolution_inner_layout.addWidget(
            self.custom_scaling_edit.textbox, 2, 2)
        screen_resolution_inner_layout.addWidget(
            self.custom_scaling_edit.help_label, 2, 3)
        screen_resolution_inner_layout.setColumnStretch(2, 1)

        screen_resolution_layout.addLayout(screen_resolution_inner_layout)
        screen_resolution_group.setLayout(screen_resolution_layout)

        if sys.platform == "darwin" and not is_conda_based_app():
            self.create_tab(
                _("Interface"),
                [screen_resolution_group, interface_group, macOS_group]
            )
        else:
            self.create_tab(
                _("Interface"),
                [screen_resolution_group, interface_group]
            )

        self.create_tab(_("Advanced settings"), advanced_widget)

    def perform_checks(self):
        # Prevent setting an empty scale factor in case users try to do it.
        # See spyder-ide/spyder#21733 for the details.
        if self.custom_scaling_radio.radiobutton.isChecked():
            scale_factor = self.custom_scaling_edit.textbox.text()
            if scale_factor == "":
                self.normal_radio.radiobutton.setChecked(True)
                self.changed_options.add('high_dpi_custom_scale_factors')

        um = self.plugin.get_plugin(Plugins.UpdateManager, error=False)
        if (
            um
            and ('update_manager', 'check_stable_only') in self.changed_options
        ):
            um.update_manager_status.set_no_status()


    def _save_lang(self):
        """
        Get selected language setting and save to language configuration file.
        """
        for combobox, (sec, opt, _default) in list(self.comboboxes.items()):
            if opt == 'interface_language':
                data = combobox.itemData(combobox.currentIndex())
                value = from_qvariant(data, to_text_string)
                break
        try:
            save_lang_conf(value)
            self.set_option('interface_language', value)
        except Exception:
            QMessageBox.critical(
                self,
                _("Error"),
                _("We're sorry but the following error occurred while trying "
                  "to set your selected language:<br><br>"
                  "<tt>{}</tt>").format(traceback.format_exc()),
                QMessageBox.Ok
            )
            return



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Main interpreter entry in Preferences."""

# Standard library imports
import os
import os.path as osp
import sys

# Third party imports
from qtpy.QtWidgets import (QButtonGroup, QGroupBox, QInputDialog, QLabel,
                            QLineEdit, QMessageBox, QPushButton, QVBoxLayout)

# Local imports
from spyder.api.translations import _
from spyder.api.preferences import PluginConfigPage
from spyder.py3compat import to_text_string
from spyder.utils import programs
from spyder.utils.conda import get_list_conda_envs_cache
from spyder.utils.misc import get_python_executable
from spyder.utils.pyenv import get_list_pyenv_envs_cache


class MainInterpreterConfigPage(PluginConfigPage):

    def __init__(self, plugin, parent):
        super().__init__(plugin, parent)
        self.apply_callback = self.perform_adjustments
        self.plugin_container = plugin.get_container()

        self.cus_exec_radio = None
        self.pyexec_edit = None
        self.cus_exec_combo = None

        conda_env = get_list_conda_envs_cache()
        pyenv_env = get_list_pyenv_envs_cache()
        envs = {**conda_env, **pyenv_env}
        valid_custom_list = self.get_option('custom_interpreters_list')
        for env in envs.keys():
            path, _ = envs[env]
            if path not in valid_custom_list:
                valid_custom_list.append(path)
        self.set_option('custom_interpreters_list', valid_custom_list)

        # add custom_interpreter to executable selection
        executable = self.get_option('executable')

        # check if the executable is valid - use Spyder's if not
        if self.get_option('default') or not osp.isfile(executable):
            executable = get_python_executable()
        elif not self.get_option('custom_interpreter'):
            self.set_option('custom_interpreter', ' ')

        self.plugin_container.add_to_custom_interpreters(executable)
        self.plugin_container.validate_custom_interpreters_list()

    def initialize(self):
        super().initialize()

    def setup_page(self):
        newcb = self.create_checkbox

        # Python executable Group
        pyexec_group = QGroupBox(_("Python interpreter"))
        pyexec_bg = QButtonGroup(pyexec_group)
        pyexec_label = QLabel(
            _(
                "Select the default Python interpreter to create new IPython "
                "consoles and to provide code completion in the Editor"
            )
        )
        pyexec_label.setWordWrap(True)
        self.def_exec_radio = self.create_radiobutton(
            _("Internal (i.e. the same used by Spyder)"),
            'default',
            button_group=pyexec_bg,
        )
        self.cus_exec_radio = self.create_radiobutton(
            _("Use the following interpreter:"),
            'custom',
            button_group=pyexec_bg,
        )

        if os.name == 'nt':
            filters = _("Executables") + " (*.exe)"
        else:
            filters = None

        pyexec_layout = QVBoxLayout()
        pyexec_layout.addWidget(pyexec_label)
        pyexec_layout.addWidget(self.def_exec_radio)
        pyexec_layout.addWidget(self.cus_exec_radio)
        self.plugin_container.validate_custom_interpreters_list()
        self.cus_exec_combo = self.create_file_combobox(
            _('Recent custom interpreters'),
            self.get_option('custom_interpreters_list'),
            'custom_interpreter',
            filters=filters,
            default_line_edit=True,
            adjust_to_contents=True,
            validate_callback=programs.is_python_interpreter,
        )
        self.cus_exec_combo.setStyleSheet("margin-left: 3px")
        self.cus_exec_combo.combobox.setMinimumWidth(400)

        self.def_exec_radio.radiobutton.toggled.connect(
            self.cus_exec_combo.setDisabled)
        self.cus_exec_radio.radiobutton.toggled.connect(
            self.cus_exec_combo.setEnabled)
        pyexec_layout.addWidget(self.cus_exec_combo)
        pyexec_group.setLayout(pyexec_layout)

        self.pyexec_edit = self.cus_exec_combo.combobox.lineEdit()

        # UMR Group
        umr_group = QGroupBox(_("User Module Reloader (UMR)"))
        umr_label = QLabel(_("UMR forces Python to reload modules which were "
                             "imported when executing a file in a Python or "
                             "IPython console with the <i>runfile</i> "
                             "function."))
        umr_label.setWordWrap(True)
        umr_enabled_box = newcb(
            _("Enable UMR"),
            'umr/enabled',
            msg_if_enabled=True,
            msg_warning=_(
                "This option will enable the User Module Reloader (UMR) "
                "in Python/IPython consoles. UMR forces Python to "
                "reload deeply modules during import when running a "
                "Python file using the Spyder's builtin function "
                "<b>runfile</b>."
                "<br><br><b>1.</b> UMR may require to restart the "
                "console in which it will be called "
                "(otherwise only newly imported modules will be "
                "reloaded when executing files)."
                "<br><br><b>2.</b> If errors occur when re-running a "
                "PyQt-based program, please check that the Qt objects "
                "are properly destroyed (e.g. you may have to use the "
                "attribute <b>Qt.WA_DeleteOnClose</b> on your main "
                "window, using the <b>setAttribute</b> method)"
            ),
        )
        umr_verbose_box = newcb(
            _("Show reloaded modules list"),
            'umr/verbose',
            msg_info=_("Please note that these changes will "
                       "be applied only to new consoles"),
        )
        umr_namelist_btn = QPushButton(
            _("Set UMR excluded (not reloaded) modules"))
        umr_namelist_btn.clicked.connect(self.set_umr_namelist)

        umr_layout = QVBoxLayout()
        umr_layout.addWidget(umr_label)
        umr_layout.addWidget(umr_enabled_box)
        umr_layout.addWidget(umr_verbose_box)
        umr_layout.addWidget(umr_namelist_btn)
        umr_group.setLayout(umr_layout)

        vlayout = QVBoxLayout()
        vlayout.addWidget(pyexec_group)
        vlayout.addWidget(umr_group)
        vlayout.addStretch(1)
        self.setLayout(vlayout)

    def warn_python_compatibility(self, pyexec):
        if not osp.isfile(pyexec):
            return

        spyder_version = sys.version_info[0]
        try:
            args = ["-c", "import sys; print(sys.version_info[0])"]
            proc = programs.run_program(pyexec, args, env={})
            console_version = int(proc.communicate()[0])
        except IOError:
            console_version = spyder_version
        except ValueError:
            return False

        if spyder_version != console_version:
            QMessageBox.warning(
                self,
                _('Warning'),
                _("You selected a <b>Python %d</b> interpreter for the "
                  "console but Spyder is running on <b>Python %d</b>!.<br><br>"
                  "Although this is possible, we recommend you to install and "
                  "run Spyder directly with your selected interpreter, to "
                  "avoid seeing false warnings and errors due to the "
                  "incompatible syntax between these two Python versions."
                  ) % (console_version, spyder_version),
                QMessageBox.Ok,
            )

        return True

    def set_umr_namelist(self):
        """Set UMR excluded modules name list"""
        arguments, valid = QInputDialog.getText(
            self,
            _('UMR'),
            _("Set the list of excluded modules as this: "
              "<i>numpy, scipy</i>"),
            QLineEdit.Normal,
            ", ".join(self.get_option('umr/namelist')),
        )
        if valid:
            arguments = to_text_string(arguments)
            if arguments:
                namelist = arguments.replace(' ', '').split(',')
                fixed_namelist = []
                non_ascii_namelist = []
                for module_name in namelist:
                    if programs.is_module_installed(module_name):
                        fixed_namelist.append(module_name)

                invalid = ", ".join(set(namelist) - set(fixed_namelist) -
                                    set(non_ascii_namelist))
                if invalid:
                    QMessageBox.warning(
                        self,
                        _('UMR'),
                        _("The following modules are not "
                          "installed on your machine:\n%s") % invalid,
                        QMessageBox.Ok,
                    )
                QMessageBox.information(
                    self,
                    _('UMR'),
                    _("Please note that these changes will "
                      "be applied only to new IPython consoles"),
                    QMessageBox.Ok,
                )
            else:
                fixed_namelist = []

            self.set_option('umr/namelist', fixed_namelist)

    def perform_adjustments(self):
        """Perform some adjustments to the page after applying preferences."""
        if not self.def_exec_radio.radiobutton.isChecked():
            # Get current executable
            executable = self.pyexec_edit.text()
            executable = osp.normpath(executable)
            if executable.endswith('pythonw.exe'):
                executable = executable.replace("pythonw.exe", "python.exe")

            # Update combobox items.
            custom_list = self.cus_exec_combo.combobox.choices
            if executable not in custom_list:
                custom_list = custom_list + [executable]
                self.cus_exec_combo.combobox.clear()
                self.cus_exec_combo.combobox.addItems(custom_list)
            self.pyexec_edit.setText(executable)

            # Show warning compatibility message.
            self.warn_python_compatibility(executable)
